/**
 * \file GeoData.hpp
 *
 * GeoData header file to declare geographical data manipulation functions
 *
 * \author Redouane Boumghar, Hung Cao
 *
 * \ingroup lgl
 *
 **/


#ifndef LGL_GEODATA_HPP
#define LGL_GEODATA_HPP

#include <list>
#include <vector>
#include <iostream>
#include <fstream>
#include <string>

#include <gdal_priv.h>
//#include <cpl_string.h>
#include <ogr_spatialref.h>

#include <lgl/RasterBand.hpp>
#include <lgl/InfoBand.hpp>
#include <lgl/LandmarkBand.hpp>
#include <lgl/Attributes.hpp>


namespace jafar {
	namespace lgl {
	 

		// the index of raster bands inside the data set 
		typedef enum {GEODATA_TFLAT_BAND=1,
		              GEODATA_TROUGH_BAND,
		              GEODATA_TRICLASSOBST_BAND,
		              GEODATA_LANDMARK_BAND,
					  GEODATA_VISIBILITY_BAND,
					  GEODATA_VISIBILITY_HEIGHT_BAND,
		              GEODATA_NMAX_BAND} GEODATA_BAND;

		typedef enum {ENTROPY_MAP=0,
		              POTENTIAL_MAP,
		              UTILITY_MAP} GEODATA_INTERNAL_MAP;

		/** Class for GDAL data handling
		 *
		 * @ingroup lgl
		 */
		class GeoData : public Raster
		{ /*{{{*/ 
			public:

				static const double DEFAULT_DATA_VALUE() { return -1.0; };
				static const double GEODATA_DEFAULT_SCALE() { return 1.0; };
			
				//enum FUSION_METHOD // defined in RasterBand.hpp

				
				/** Default Constructor with no image or data load */
				GeoData();

				/** Constructor used to load a map from a raster file */
				GeoData(const std::string& _dsfilename, FUSION_METHOD _fmethod=MEAN);

				/** Constructor used to load a map from buffered data */
				GeoData(int _xsize, int _ysize, double* _bufferedData, FUSION_METHOD _fmethod=MEAN);

				/** Destructor */
				~GeoData();

				bool loadBand(GEODATA_BAND band, const std::string& _dsfilename);
				bool loadInfoBand(const std::string& _dsfilename, GEODATA_BAND aband);
				bool loadLandmarkBand(const std::string& _dsfilename);

				/** Virtual file detection flag
				 *
				 * This option is used to enable detection of virtual memory Dataset
				 * on which FlushCache is impossible.
				 *
				 *  0 - classical opening of a file on a hard memory device
				 *  1 - file created from a buffer with a file created on the virtual memory
				 */
				int isInVirtualMemory;

				/** Array to hold the goal reaching potential information */
				double* mapPotential;

				/** Array to hold the information entropy */
				double* mapH;

				double* mapUtility;
				//------------------- getters

				GDALDataset* getDataset() const;

				void getAffineTransform(double& _utmXRoot, double& _utmYRoot, double& _cellXRes, double& _cellYRes); 

				/**
				 * Abstract methods from abstract class Raster 
				 */			
				int getxroot() const;
				int getyroot() const;
				int getxsize() const;
				int getysize() const;

				double getScale() const;

				/** Geographical X coordinate of the top left corner */
				double geoxroot() const;
				/** Geographical Y coordinate of the top left corner */
				double geoyroot() const;
				/** Geographical X coordinate of the down right corner */
				double geoxend() const;
				/** Geographical Y coordinate of the down right corner */
				double geoyend() const;
				/** Geographical X resolution of raster pixels */
				double geoxres() const;
				/** Geographical Y resolution of raster pixels */
				double geoyres() const;

				/** Get the value loaded from the band _aband at given index */
				double getValueFromBand(GEODATA_BAND _aband, int index);

				/** Get the value loaded from the band _aband at given x and y
				 * coordinates of the top left pixel corner
				 */
				double getValueFromBand(GEODATA_BAND _aband, int xindex, int yindex);

				// get the value atr given raster cell (x,y)
				double getData(GEODATA_BAND band, const RasterCellIndex& cell) const;

				// fusion
				double getData(const RasterCellIndex& cell) const;
				
				// get the fusion value of all bands at cluster rect
				double getData(const RasterRect& rect) const;
				double getData(GEODATA_BAND band, const RasterRect& rect) const;
				
				GDRasterBand<double>* getBand(GEODATA_BAND band);
				
				double getInterpolatedValue(GEODATA_BAND _aband, double xNavLoc, double yNavLoc);
				//-------------- setters

				/** Set the GDAL dataset scale */
				void setScale(double _scale);

				/** Set/Updates the value in band _aband at index _index using
				 * _val as the new value */
				void setValueInBand(GEODATA_BAND _aband, int _index, double _val);

				/** Set the fusing method to an available fusing flag 
				 *
				 * @see fmethod
				 */
				void setFusingMethod(FUSION_METHOD _fmethod);

				//void setGeoTransform(double _utmXRoot, double _utmYRoot, double _cellXRes, double _cellYRes); 
				bool setData(double* bufferedData, double* bufferedOther, int _xsize, int _ysize, int _band = GEODATA_TFLAT_BAND);

				/** Write attributes in cell covering position (at_i, at_j)
				 *
				 * It can be called with internalTransfer = false in order to
				 * avoid multiple calls of memToBuffer in case setData is used
				 * in a loop. This greatly fasten the process.
				 *
				 * @param at_i  i coordinate of the set pixel.
				 * @param at_j  j coordinate of the set pixel.
				 * @param inatt  Attributes, information input to set.
				 * @param internalTransfer Default is true to enable the memToBuffer transfer.
				 */
				bool setData(int at_i, int at_j, Attributes inatt, bool internalTransfer=true);

				/** Modifies the decomposition limit on all bands */
				void setDecompositionLimit(double _dl);

				/** Set decomposition type for all bands */
				void setDecompositionType(HOMOGENEOUS_TYPE ht_);

				/** Set the probability lag used when the decomposition type is NEARLY */
				void setProbabilityLag(double probaLag_);

				/** Set the sigmaf factor used while the decomposition type is SIGMA */
				void setSigmaFactor(double sigmaF_);

				/** Set the simulationProbaVisionLimit factor */
				void setSimulationProbaVisionLimit(double spvl_);

				//-------------- testers
				/**
				 * given a rect ((top,left),(right,bottom)), tell if this is homogeneous
				 */
				bool isHomogeneous(const RasterRect& rect) const;

				// same value at all bands
				bool sameValue(const RasterRect& rect);
				// same value at given band
				bool sameValue(GEODATA_BAND band, const RasterRect& rect);

				/** Verify if geodata is overlapping at least with 1 pixel with _toMerge geodata 
				 * @return true if overlapping
				 */
				bool overlaps(GeoData* _toMerge);

				//---------------- smarters

				/** Merge with another geodata.
				 *
				 * The Common overlapping region is calculated and the fusing method is called on this region.
				 * The data is fused in "this" and its size is not changed so there is merging
				 * only if "this" overlaps with "_toMerge".
				 * Fusion is made according to the choosen fusion method (fmethod)
				 *
				 * @param _toMerge GeoData to merge with the current data
				 * @param _aband The band that is to be merged
				 * @see fmethod
				 */
				bool mergeStrict(GeoData* _toMerge, GEODATA_BAND _aband);

				/** Add information to the GDAL Dataset
				 * @param _dataset GDAL Dataset to fuse
				 * @see mergeStrict()
				 */
				void addDataset(GDALDataset *_dataset);

				/** Fusing information from the fusingRect region from _geodata
				 * into innerRect region of the current geodata. In this method
				 * we consider that the fusingRect is always strictly included
				 * in the boundaries of the current GeoData and is equal in
				 * real geographic region to the one defined by innerRect
				 *
				 * @param _geodata GeoData to fuse with.
				 * @param _aband Band number to fuse.
				 * @param innerRect Zone in which data have to be fused
				 * @param fusingRect Zone from which data will be fused
				 */
				void fuseData(GeoData *_geodata2, GEODATA_BAND _aband, const RasterRect& innerRect, const RasterRect& fusingRect);
				
				/** Calculate bands homogeneity percentage
				 *
				 * It is the rate of pixel with differing direct-neighborhood
				 * to the total number of considered pixels (image borders are
				 * ignored).
				 */
				double calculateBandHomogeneityPercentage(GEODATA_BAND _aband);

				/** Print useful information on the loaded geo data */
				void printInfo(bool verbose = true) const;

				/** Saving images in GeoTif format by default */
				bool saveData(const std::string& outfile, const char* wFormat="GTIFF");

				/** Transport info from GDAL Datasets to Bands' internal data */
				void bufferToMem();

				/** Transport info from Bands' internal data to GDAL Datasets */
				void memToBuffer();

				/** Saving only traversability layer in a ppm file */
				bool saveTraversability(const std::string& outfile);

				/** Reset knowledge on all layers */
				void resetBands();

				/** Add a new information band to the dataset 
				 *
				 * TODO: dynamic array of bands to extend is needed
				 *
				 * For now prints if ading a new GDT_Float64 band is possible
				 */
				bool createNewInfoBand(GDInfoBand** thatNewBand);

				/** Initialized or reset utility map with zeros */
				bool initUtility();

				/** Sets the goal reaching potential from (x1_, y1_) to (x2_, y2_) 
				 *
				 * Data is interpolated and stored in the corresponding map 
				 * for all points between  (x1_, y1) and (x2_, y2_)
				 *
				 * @param v1_ goal reaching potential value of closest-to-the-goal node
				 * @param x1_ x coord of node with potential v1_
				 * @param y1_ y coord of node with potential v1_
				 * @param v2_ goal reaching potential value of previous node in potential path
				 * @param x2_ x coord of node with potential v2_
				 * @param y2_ y coord of node with potential v2_
				 * @see buildUtility() initUtility()
				 */
				bool setMap(GEODATA_INTERNAL_MAP map2set,double v1_, int x1_, int y1_, double v2_, int x2_, int y2_);

				/** Merge potential with entropy to construct navigation utility */
				bool buildUtility();

				void resetUtility();

				/** Write a ppm image of the utility map : mapUtility */
				bool saveUtility(const std::string& outfile);
				/** Write a ppm image of the entropy map : mapH */
				bool saveEntropy(const std::string& outfile);
				/** Write a ppm image of the potential map : mapPotential */
				bool savePotential(const std::string& outfile);

				/** Returns the planar euclidean distances between two points */
				double euclideanDistance(double, double, double, double);

				bool setUTMProjection(GDALDataset* inputDataset, int utmZone);

				/** Getting mean entropy in a sensor foot print */
				double getFootPrintEntropy(int, int);

				/** Returns the decomposition type of this band */
				HOMOGENEOUS_TYPE getTraversabilityHomogeneousType();
				/** Returns the decomposition type of this band */
				HOMOGENEOUS_TYPE getRoughHomogeneousType();
				/** Get the probability lag used for decomposition in the traversability band */
				double getTraversabilityProbabilityLag();
				/** Get the probability lag used for decomposition in the rough terrain band */
				double getRoughProbabilityLag();
				/** Get the sigma used for decomposition in the traversability band */
				double getTraversabilitySigmaFactor();
				/** Get the sigma used for decomposition in the rough terrain band */
				double getRoughSigmaFactor();
				/** Get the simulationProbaVisionLimit factor */
				double getSimulationProbaVisionLimit();


				/** Tell if the land is traversable by the line of vision
				 *
				 * @see simulationProbaVisionLimit
				 */
				bool isVisionTransparent(int atx, int aty, int atz=0);

				/** Load the visibility Data from a file pointed by _visibilityFileName
	 			 *
				 * @param _dsfilename File Path to open ; it has to be a visibility map
				 *  
				 */
				bool loadVisibility ( const std::string& _visibilityFileName );

			private:

				/** Quadtree maximum cell size in meter */
				double decompositionLimit;

				/** Loaded file name */
				std::string dsfilename;

				/** GDALDataset of the loaded image or any other GDAL Compatible file
				 * This holds every possible information about it.
				 */
				GDALDataset* dataset;		

				/** Coefficient of affine Geo reference transformation array:
				 * [0] : X origin of left top corner of left top pixel
				 * [3] : Y origin of ...
				 * [1] : Width in meters of this pixel
				 * [5] : Height in meters of this pixel
				 * [2] : Lines offset/shifts/drift
				 * [4] : Columns offset/shifts/drift
				 */
				double refGeoTransform[6];

				/** Return type of dataset->GetGeoTransform() function that
				 * fills the refGeoTransform array */
				CPLErr refGeoTransformRType;

				//------------- GDRasterBand s -------------//

				/** Main band is the combination of other information bands
				 * (like trav and landmark) */
				//GDRasterBand<double>* mainBand;

				/** Traversability Band */
				GDInfoBand* traversabilityBand;
				/** Landmark Band */
				GDLandmarkBand* landmarkBand;
				/** Rough Terrain Band */
				GDInfoBand* roughTerrainBand;
				/** Visibility Band */
				GDInfoBand* visibilityBand;
				/** Visibility Height Band */
				GDInfoBand* visibilityHeightBand;

				/** Seems unused for now.
				 *
				 * Should be the fusion methods between bands. Used to produce the main band.
				 * The are defined in RasterBand.hpp
				 */
				FUSION_METHOD fmethod;

				/** Limit used to evaluate intervisibility on real worlds since
				 * on simulated world everything is fully state-defined.
				 *
				 * Terrain with a sum of P(all-traversable-classes) <  simulationProbaVisionLimit
				 * are not transparent to vision.
				 */
				double simulationProbaVisionLimit;

				/** Prints information about the GDal Dataset 
				 * This method is called in the redifinition of operator "<<"
				 */
				void info(std::ostream& info, bool verbose = true) const;
						
				friend std::ostream& operator<< (std::ostream& out, const GeoData& gd);

				/** Method to load an image or use an existing dataset
				 *
				 * When loading GeoData from a file the file should at least
				 * contains one band of information or contain 3 channels.
				 * - 8-Bit One band images are greyscale images: The band is used
				 * as the traversability band and obstacle value is
				 * completarily deducted.
				 * - RGB images : Red channel is not used
				 *   Green Channel is used for flat terrain probability (traversabilityBand)
				 *   Blue Channel is used of rough terrain probability (roughTerrainBand)
				 *   The obstacle probability is then complementarily deducted.
				 *
				 * @param _dsfilename File path of the file to open
				 * @param _useThisDataset If not NULL Take precedence on _dsfilename 
				 * @return true  When the file could be processed  (else false).
				 */
				bool loadData(const std::string& _dsfilename, GDALDataset* _useThisDataset=NULL);

				/**
				 * Load visibility GeoData from a file
	 			 *
	 			 * @param _visibilityFileName File path of the file to open
	 			 * @return true  When the file could be processed  (else false).
	 			 */
				bool loadVisibilityData(const std::string& _visibilityFileName, GDALDataset* _useThisDataset);

				// Friendly class
				friend class NavGraph;
				friend class PositionMap;
				friend class Decomp; 

		}; /*}}} end of class GeoData*/

		/** 
		 * Redefinition of operator <<
		 */
		inline std::ostream& operator<< (std::ostream& out, const GeoData& gd) {
			gd.info(out,true);
			return out;
		}
	} // namespace lgl
} // namespace jafar

#endif /* LGL_GEODATA_HPP */
