/**
 * \file LGLViewer.hpp
 *
 * Header file to declare Qt drawing objects and functions
 *
 * \author Redouane Boumghar, Hung Cao
 *
 * \ingroup lgl
 *
 **/

/* $Id: lglViewer.hpp 4094 2009-09-04 09:32:58Z hcao $ */

#ifdef HAVE_MODULE_QDISPLAY // lgl viewer is on if HAVE_MODULE_QDISPLAY is defined

#ifndef LGL_DISPLAY_HPP
#define LGL_DISPLAY_HPP

#include <boost/graph/iteration_macros.hpp>
#include <ostream>

#include <qdisplay/PolyLine.hpp>
#include <qdisplay/Viewer.hpp>
#include <qdisplay/ViewerManager.hpp>
#include <qdisplay/ImageView.hpp>
#include <qdisplay/Line.hpp>
#include <qdisplay/Shape.hpp>
#include <qdisplay/ImageView.hpp>
#include <qdisplay/AbstractEventHandler.hpp>
#include <qdisplay/Ellipsoid.hpp>
#include <qdisplay/init.hpp>
#include <qdisplay/Viewer.hpp>

#include <lgl/NavGraph.hpp>
#include <lgl/VisibilityMap.hpp>
#include <lgl/VisibleArea.hpp>

namespace jafar {
	namespace lgl {
		
		/** Maximum number of paths that we allow to be drawn */
		#define NMAX_PATHS 30

		/** Node states enumeration */
		typedef enum {
			/** State of node when in the A* open list */
			DRAW_NODE_OPEN=1,
			/** State when popped  out the open list */
			DRAW_NODE_VISITED,
			/** State when in the A* path */
			DRAW_NODE_PATH,
			/** State when in the robot executed path */
			DRAW_NODE_TRACK
			} DRAW_NODE_STATE;

		class LGLViewer {
			public:
				
				LGLViewer();

				/** Smart Constructor to tell which quadtree and graph to visualize 
				 *
				 * @param _decomp  Decomposition taken from one of the quadtree decomposition level.
				 * @param _graph   Graph that has been associated to this decomposition or another.
				 * @param sf       Linear scale factor to increase displayed image's size.
				 */
				LGLViewer(Decomp* _decomp, NavGraph::NavGraphContainer* _graph, double sf=1.0);

				/** Constructor for visibility  
				 *
				 * @param _visibility 	VisibilityMap (visibility must have been compute for useful display).
				 * @param _decomp  		Decomposition taken from one of the quadtree decomposition level.
				 * @param _graph   		Graph that has been associated to this decomposition or another.
				 * @param sf       		Linear scale factor to increase displayed image's size.
				 */
				LGLViewer(VisibilityMap *_visibility, Decomp* _decomp=NULL, NavGraph::NavGraphContainer* _graph=NULL, double sf=1.0);
				
				~LGLViewer();

				jafar::qdisplay::Viewer* getQViewer();

				/** Displays the Quadtree Decomposition. It can be superimposed with the graph using displayGraph */
				bool displayDecomp();

				/** Display the graph superimposable with the quadtree decomposition displayed using displayDecomp() */
				bool displayGraph(bool show_edges=true, bool show_nodes=true); 

				/** Display the visibility Map for a given node */
				bool displayVisibilityMap(const int x, const int y, bool show_viewlines=false, bool show_nodes=true, bool show_rays=false); 

				/** Display a given visible Area
				 * (NB: a visibilityMap must have been set
				 */
				bool displayVisibleArea(const VisibleArea &myArea, bool show_viewlines=false, bool show_nodes=true, bool show_rays=false); 

				/** Display the graph using its UTM coordinates */
				bool displayGraphWithUTM(bool show_edges=true, bool show_nodes=true); 

				/** Draw a simple line between two given nodes */
				void drawEdge(const NavGNode& prevNode, const NavGNode& curNode, DRAW_NODE_STATE dstate);

				/** Draw a simple line between two given position */
				void drawEdge(const NavLocation& prevPos, const NavLocation& curPos, DRAW_NODE_STATE dstate);

				/** Draw lines and nodes between list of waypoints */
				void displayPath(NavPath& dPath);

				/** Display image "filename" in the viewer window */
				void setImage(const std::string& filename, double xoffset=0.0, double yoffset=0.0);

				/** Pointer function to send to boost::astar visitor in order to watch whats happening */
				void displayHook(const NavGNode& u, DRAW_NODE_STATE dstate);
				void displayHook(const NavLocation& pos, DRAW_NODE_STATE dstate);
				
				//---------- setters
				void setDecomp(Decomp*);
				void setGraph(NavGraph::NavGraphContainer*);
				void setScaleFactor(double);

			private:

				/** True if we are using UTM position to display graph objects */
				bool withUTM;

				/** Pointer to the decomposition object */
				Decomp* decomp;
				
				/** Graph container */
				NavGraph::NavGraphContainer* graph;

				/** Scale factor to effect display
				 *  If > 1 the object will be sized up.
				 */
				double scaleFactor;

				/** QDisplay viewer used to draw shapes */
				jafar::qdisplay::Viewer* viewer;

				/** Number of path items drawn 
				 *
				 * Used to keep track of the number of path and to change colors generically
				 */
				int n_path_drawn;

				/** VisibilityMap, use to display visibility relationships
				 */
				VisibilityMap *visibility;	

			
		};
	}
}

#endif // LGL_DISPLAY_HPP
#endif // QDisplay enabled 
