/**
 * \file Raster.hpp
 *
 * Header file for the declaration and definition of all raster small objects
 *
 * 
 *
 * \author Redouane Boumghar, Hung Cao
 *
 * \ingroup lgl
 *
 **/

#ifndef LGL_RASTER_HPP
#define LGL_RASTER_HPP

#include <iostream>
#include <algorithm>


namespace jafar {
	namespace lgl {
 
 		/** Point object in a Raster. 
		 * It defines the top left corner of a unit cell, pixel.
		 */
		struct RasterPoint {
			int x; 
			int y;

			RasterPoint():x(0),y(0) {}
			RasterPoint(int _x,int _y):x(_x),y(_y) {}

			void set(int _x, int _y) {x=_x;y=_y;}

			bool operator==(const  RasterPoint &rhs) const
			{
				return (x == rhs.x) && (y == rhs.y);
			}  

			bool operator!=(const  RasterPoint &rhs) const
			{
				return (x != rhs.x) || (y != rhs.y);
			}  

		};

		/** Top left corner coordinates in 2D of a Cell (= a pixel or a cluster of pixels). */
		struct RasterCellIndex {

			int i; 
			int j;

			RasterCellIndex():i(0),j(0) {}
			RasterCellIndex(int _i,int _j):i(_i),j(_j) {}

			void set(int _i, int _j) {i=_i;j=_j;}

			bool operator==(const  RasterCellIndex &rhs) const
			{
				return (i == rhs.i) && (j == rhs.j);
			}  

			bool operator!=(const  RasterCellIndex &rhs) const
			{
				return (i != rhs.i) || (j != rhs.j);
			}  

		};

		std::ostream& operator<< (std::ostream& out, const RasterCellIndex& cell);

		/** RasterRect class
		 *
		 * This object is used to deal with clusters of pixels.
		 * Its coordinates are pointed by the top left corner of its TOP-LEFT pixel
		 * and by the coordinates of the top-left corner of the BOTTOM-RIGHT pixel.
		 *
		 * To correspond to GDAL Raster system the coordinates frame is the following :
		 *  +----------- (+ x) ------>
		 *  |
		 *  |
		 *  |
		 *  |
		 * (+y)
		 *  |
		 *  |
		 *  V
		 */
		class RasterRect { /* {{{ */
			public:

				RasterRect();
				RasterRect( int _xroot, int _yroot, int _xsize, int _ysize );

				/** Valid rectangles return true if the bottom right corner is
				 * really at the bottom right or on the top left corner or at
				 * least the same. */ 
				bool  isValid() const;

				int &rLeft();
				int &rTop();
				int &rRight();
				int &rBottom();

				// Next functions do the same as left() top() right() and bottom()
				/** Get the X coordinate of the top left corner of the top left
				 * pixel/cell of the current rectangle */
				int   xroot()const;
				/** Get the Y coordinate of the top left corner of the top left
				 * pixel/cell of the current rectangle */
				int   yroot()const;
				/** Get the X coordinate of the top left corner of the bottom
				 * right pixel/cell of the current rectangle
				 *
				 * One should use (xend() + 1) to get the coordinate of the
				 * bottom right corner of the full rectangle
				 */
				int   xend()const;
				/** Get the Y coordinate of the top left corner of the bottom
				 * right pixel/cell of the current rectangle 
				 *
				 * One should use (yend() + 1) to get the coordinate of the
				 * bottom right corner of the full rectangle
				 */
				int   yend()const;

				void   setXTopLeft( int pos );
				void   setYTopLeft( int pos );
				void   setXBottomRight( int pos );
				void   setYBottomRight( int pos );
				void   setX( int x );
				void   setY( int y );

				RasterPoint topLeft() const;
				RasterPoint bottomRight() const;
				RasterPoint topRight() const;
				RasterPoint bottomLeft() const;
				/** The center is calculated in integer values */
				RasterPoint center() const;

				void rect( int& _xtopleft, int& _ytopleft, int& _xsize, int& _ysize ) const;
				void coords( int& xtopleft, int& ytopleft, int& xbottomright, int& ybottomright ) const;

				void setRect( int x, int y, int w, int h );
				void setCoords( int xtopleft, int ytopleft, int xbottomright, int ybottomright );

				/** Returns the total number of pixels inside the rectangle */
				int size()const;

				/** Returns xsize 
				 * @sa xsize 
				 */
				int xsize() const { return xpsize; }

				/** Returns ysize 
				 * @sa ysize 
				 */
				int ysize() const { return ypsize; }

				RasterRect&  operator=(const RasterRect &rect);

				/** This function tells (true) if the given cell index is part of the current raster rectangle (R.O.I.)
				 * Raster cells indexes are given as the top left coordinates of a cell. So fully contained cells cannot have
				 * coordinates equal to right and bottom sides of the container.
				 *
				 * @param cell RasterCellIndex to be checked if contained by the current rectangle
				 * @param inside  if true only fully contained cells are
				 * considered else cells origin contained in the rectangle and
				 * its borders are considered
				 * @return true if RasterCellIndex cell is included in this(RasterRect)
				 */
				bool contains( const RasterCellIndex& cell, bool inside=true) const;

				/** This function tells (true) if the given RasterRect is part of the current raster rectangle (R.O.I.)
				 *
				 * @param r RasterRect to be checked if contained by the current rectangle
				 * @param inside  if true only fully contained rects are
				 * considered else rects origin contained in the rectangle and
				 * its borders are considered
				 * @return true if RasterRect r is included in this(RasterRect)
				 */
				bool contains( const RasterRect &r, bool inside=true ) const;

				/** Test RasterRect intersection and creates resulting intersection */
				bool intersects( const RasterRect &r, RasterRect &result ) const;

				friend bool operator==( const RasterRect &, const RasterRect & );
				friend bool operator!=( const RasterRect &, const RasterRect & );
				friend std::ostream& operator<<(std::ostream&, const RasterRect &);
				  
				/** Top left corner X coordinate in raster cell number */
				int xtopleft;
				/** Top left corner Y coordinate in raster cell number */
				int ytopleft;
				/** Bottom right corner X coordinate in raster cell number */
				int xbottomright;
				/** Bottom right corner Y coordinate in raster cell number */
				int ybottomright;

				/** Size in pixel units on X axis */
				int xpsize;
				/** Size in pixel units on Y axis */
				int ypsize;

		}; /* }}} */

		bool operator==( const RasterRect &, const RasterRect & );
		bool operator!=( const RasterRect &, const RasterRect & );

		/**
		* Abtract class to define the type suitable for quadtree representation 
		*/
		class Raster  { /*{{{*/
			public:

				enum CLUSTERING_FUNCTION {CF_EQUALITY, CF_STD_DEV};

				Raster(CLUSTERING_FUNCTION cf=CF_EQUALITY):
				  clusteringF(cf)
				{}

				virtual ~Raster() {}

				void setClusteringF(CLUSTERING_FUNCTION cf);
				CLUSTERING_FUNCTION getClusteringF();

				void setClusteringThres(double _cfthreshold);
				double getClusteringThres();

				RasterRect getRasterRect() const;

				/** Returns true if rect is totally inside the current Raster */
				bool inside(const RasterRect& rect);

				/** RasterCellIndex must be in raster boundaries */
				bool validIndex(const RasterCellIndex& index) const;

				/* do not forget =0 in virtual method */
				/****
				virtual int getxroot() const = 0;
				virtual int getyroot() const = 0;
				virtual int getysize() const = 0;
				virtual int getxsize() const = 0;

				// get the scale of the raster 
				virtual double getScale() const=0;
				virtual void setScale(double _scale)=0;
				*****/

				/**
				 * given a rect ((top,left),(right,bottom)), tell if this is homogeneous
				 */
				//virtual bool isHomogeneous(const RasterRect& rect) const = 0;		

			protected:

				CLUSTERING_FUNCTION clusteringF;
				double  cfthreshold;

				int xroot;
				int yroot;
				int xsize;
				int ysize;



		}; /*}}}*/
	}
}

#endif /* LGL_RASTER_HPP */
