/**
 * \file CostManager.cpp
 *
 * Source file that defines path cost functions
 *
 * \author Redouane Boumghar, Hung Cao
 *
 * \ingroup lgl
 *
 **/

#include <lgl/CostManager.hpp>

using namespace jafar::lgl;
	
CostManager::CostManager(GeoData* _geodata,  Quadtree<double,GeoData>* _quadtree, PositionManager* _posManager, DIST_TYPE _distType) {/*{{{*/
	geodata    = _geodata;
	distType   = _distType;
	quadtree   = _quadtree;
	posManager = _posManager;
}/*}}}*/
	
CostManager::~CostManager() {}
	
void CostManager::setDistType(DIST_TYPE _distType) {/*{{{*/
	distType=_distType;
}/*}}}*/

void CostManager::setObstacleLimit(double pol) {//{{{
	if (pol < ATT_MIN_VAL || pol > (1.0-ATT_MIN_VAL))
	{
		// Bad input value
		obstacleLimit = (1.0-ATT_MIN_VAL);
	} else {
		obstacleLimit = pol;
	}
}//}}}

// New proposed approach
bool CostManager::weight(RasterCellIndex locClusterRootIndex, const NavLocation& loc1, const NavLocation& loc2, double& weight) { //{{{
	DecompCluster* dcluster;
	posManager->getCluster(locClusterRootIndex, &dcluster);

	double distance = 0.0;
	switch(distType)
	{
		default:
		distance = Euclidean2DDist<NavLocation,double>(loc1,loc2);			
	};

	double navCost  = 0.0;
	double tmp = 0.0;
	int katt = 0;
	int ki = 0;
	// Building the proabilities weight array
	double wp[T_LIMIT+1];
	double wpsum  = 1.0; // Must begin at 1 to avoid minimising heuristic
	double psum   = 0.0;
	double psum_wrest   = 0.0;
	for (katt = T_FREE; katt <= T_LIMIT; katt++)
	{
		wp[katt] = 1.0 + 1000.0*(katt-T_FREE);
		wpsum = wpsum + wp[katt];
	}

	// The cost function
	for (katt = T_FREE; katt <= T_LIMIT; katt++)
	{
		psum_wrest = 0.0;
		for (ki = katt; ki <= T_LIMIT; ki++)
		{
			psum_wrest = psum_wrest + wp[ki];
		}

		psum = psum + dcluster->getCertainty((attributeKey)katt) * psum_wrest;
	}

	navCost = wpsum - psum;
	tmp = dcluster->getCertainty(T_OBSTACLE);
	if (tmp <= (obstacleLimit - ATT_MIN_VAL))
	{
		if (tmp > ATT_MIN_VAL)
		{
			//--- Constructing polynomial obstacle effect
			tmp = tmp/obstacleLimit;
			tmp = tmp * tmp; // To have a small impact on small values
			// To have a great impact on values closing to obstacleLimit
			tmp = 1.0 - tmp;
			tmp = tmp*tmp*tmp*tmp;
			// tmp is ready to be used in a logarithm form to have large impact
			// sooner than in a inverse form:
			tmp = log(tmp);
			tmp = 1.0 + tmp*tmp;
		} else {
			// Obstacle probability has no influence (zero case)
			tmp = 1.0;
		}
		navCost = navCost * tmp;
	} else {
		navCost = navCost / ATT_MIN_VAL;
	}
	weight = distance * navCost;

	return true;
} //}}}

bool CostManager::weight(const NavLocation& loc1, const NavLocation& loc2, double& weight) { //{{{
	RasterRect cluster;
	if (!posManager->getCluster(loc1,loc2,cluster)) {
		// Two possible cases:
		// - loc1 and loc2 are the same NavLocation
		// - loc1 and loc2 are not in the same cluster
		// Anyway this should be called on a well built graph and using locs of
		// the same list built thanks to getInOutPoints in order that we never
		// get inside this IF
		weight = -1.0; // throws exception "negative edge" in the astar_search
		return false;		
	}

	double distance = Euclidean2DDist<NavLocation,double>(loc1,loc2);			
	// Get : ponder(FREE)  * P(FREE)
	//     + ponder(ROUGH) * P(ROUGH)
	// ... These weighting are arbitrary. They are only useful if we
	//     artificially want to give more importance to one or another class of
	//     terrain in the cost function. For now they are all one.
	//     So here we only get the sum of probabilities what we consider NOT-OBSTACLE.
	double geoValue = quadtree->getData(cluster);

	// Weight should be light on short path that easily drivable
	// so what reduces the weight is :
	// - shortness of the path
	// - ride facility of the path
	if (geoValue > ATT_MIN_VAL)
	{
		weight = distance / geoValue;
	} else {
		weight = distance / ATT_MIN_VAL;
	}
	
	return true;
} //}}}

bool CostManager::weight(const NavLocation& loc1, const NavLocation& loc2, double costAtLoc2, double& weight) { /*{{{*/
	RasterRect cluster;
	if (!posManager->getCluster(loc1,loc2,cluster)) {
		// Two possible cases:
		// - loc1 and loc2 are the same NavLocation
		// - loc1 and loc2 are not in the same cluster
		// Anyway this should be called on a well built graph and using locs of
		// the same list built thanks to getInOutPoints in order that we never
		// get inside this IF
		weight = -1.0; // throws exception "negative edge" in the astar_search
		return false;		
	}

	double distance = Euclidean2DDist<NavLocation,double>(loc1,loc2);			
	// Get : ponder(FREE)  * P(FREE)
	//     + ponder(ROUGH) * P(ROUGH)
	// ... These weighting are arbitrary. They are only useful if we
	//     artificially want to give more importance to one or another class of
	//     terrain in the cost function. For now they are all one.
	//     So here we only get the sum of probabilities what we consider NOT-OBSTACLE.
	double geoValue = quadtree->getData(cluster);

	// costAtLoc2 is the navigation cost at the target node
	// and it is 1 for fully traversable land and max for fully obstacle land.
	//weight = costAtLoc2 * distance;
	
	// Weight should be light on short path that easily drivable
	// so what reduces the weight is :
	// - shortness of the path
	// - ride facility of the path
	if (geoValue > ATT_MIN_VAL)
	{
		//weight = (distance*costAtLoc2) / geoValue;
		weight = (distance* costAtLoc2) + distance / geoValue;
	} else {
		//weight = (distance*costAtLoc2) /ATT_MIN_VAL;
		weight =  (distance* costAtLoc2) + (distance) /ATT_MIN_VAL;
	}
	
	return true;
} /*}}}*/
