/**
 * \file GeoData.cpp
 *
 * Source code for the GeoData Class
 *
 * \date 2010-08-01
 * \author Redouane Boumghar, Hung Cao
 *
 * \ingroup lgl
 **/


#include <lgl/GeoData.hpp>
#include <kernel/jafarMacro.hpp>


namespace jafar {
	namespace lgl {

	 /*------- Constructors --------*/

	 /** Creates a GeoData object from a file pointed by _dsfilename
	  *
	  * @param _dsfilename File Path to open, it could be a raster file for traversability/landmark map,
	  *  
	  */
	GeoData::GeoData():
		dataset(NULL), 
		traversabilityBand(NULL) , 
		landmarkBand(NULL),
		roughTerrainBand(NULL),
		visibilityBand(NULL),
		visibilityHeightBand(NULL),
		fmethod(MEAN),
		simulationProbaVisionLimit(0.10)
	{/*{{{*/
		isInVirtualMemory = 0;
		// Register all GDAL Drivers
		GDALAllRegister();

		// 8 meters used for entropy interpolation
		decompositionLimit = 8.0;
		mapUtility = NULL;
		mapPotential = NULL;
		mapH = NULL;
		initUtility();
	 }/*}}}*/

	 /** Creates a GeoData object from a file pointed by _dsfilename
	  *
	  * @param _dsfilename File Path to open, it could be a raster file for traversability/landmark map,
	  *  
	  */
	GeoData::GeoData(const std::string& _dsfilename, FUSION_METHOD _fmethod):
		dsfilename(_dsfilename),
		dataset(NULL), 
		traversabilityBand(NULL) , 
		landmarkBand(NULL),
		roughTerrainBand(NULL),
		visibilityBand(NULL),
		visibilityHeightBand(NULL),
		fmethod(_fmethod),
		simulationProbaVisionLimit(0.10)
	{/*{{{*/
		isInVirtualMemory = 0;
		// Register all GDAL Drivers
		GDALAllRegister();

		if (dsfilename.empty())
		{
			std::cout << "#EEE# [GeoData] Empty file path." << std::endl;
		} else {
			loadData(dsfilename);
		}

		// 8 meters used for entropy interpolation
		decompositionLimit = 8.0;
		mapUtility = NULL;
		mapPotential = NULL;
		mapH = NULL;
		initUtility();
	 }/*}}}*/

	GeoData::GeoData(int _xsize, int _ysize, double* bufferedData, FUSION_METHOD _fmethod):
		dsfilename(""),
		dataset(NULL), 
		traversabilityBand(NULL) , 
		landmarkBand(NULL),
		roughTerrainBand(NULL),
		visibilityBand(NULL),
		visibilityHeightBand(NULL),
		fmethod(_fmethod),
		simulationProbaVisionLimit(0.10)
	{/*{{{*/
		if (_xsize == 0 || _ysize == 0)
		{
			// No suitable GDAL dataset can be created so stop here.
			std::cerr << "#EEE# Geodata : no creation possible for zero sized data." << std::endl;
			return;
		}
		isInVirtualMemory = 1;
		unsigned char* xyRaster = NULL;
		unsigned char* xyRasterBlack = NULL;
		const char *wFormat = "GTIFF";
		GDALDataset *newInputDataset = NULL;       
		GDALDriver *newInputDriver = NULL;

		// Register all GDAL Drivers
		GDALAllRegister();
		newInputDriver  = GetGDALDriverManager()->GetDriverByName(wFormat);
		if ( newInputDriver != NULL )
		{
			// Method that prints all the available option for the selected driver
			//std::cout << GDALGetDriverCreationOptionList(newInputDriver) << std::endl;
			// Converting input data [0.0 1.0] into a byte band [0 255]
			if (bufferedData != NULL)
			{
				xyRaster = (unsigned char*) calloc(_xsize*_ysize, sizeof(unsigned char));
				xyRasterBlack = (unsigned char*) calloc(_xsize*_ysize, sizeof(unsigned char));
				for (int i= 0; i < _xsize*_ysize; i++)
				{
					xyRaster[i] = (unsigned char)(bufferedData[i]*255.0);
					xyRasterBlack[i] = (unsigned char)(0);
				}
			}
		
			dsfilename = "/vsimem/lgl_initGeodataFromBuffer.tfw";
			newInputDataset = newInputDriver->Create( dsfilename.c_str(), _xsize, _ysize, 3, GDT_Byte, NULL /* papszOptions */);
		} else {
			std::cerr << "#EEE# GeoData: During object creation. No GDAL Driver found : " << wFormat << std::endl;
		}

		if ( newInputDataset == NULL )
		{
			std::cerr << "#EEE# GeoData: During object creation. Could not create Dataset with driver : " << wFormat << std::endl;
			xsize = 0;
			ysize = 0;
			xroot = 0;
			yroot = 0;
		} else {
			refGeoTransform[0] = 0.0; //Coord
			refGeoTransform[1] = 1.0; // Width
			refGeoTransform[2] = 0.0;
			refGeoTransform[3] = 0.0; //Coord
			refGeoTransform[4] = 0.0;
			refGeoTransform[5] = 1.0; // Height (could be -1.0 in some case)
			newInputDataset->SetGeoTransform( refGeoTransform );

			setUTMProjection(newInputDataset, 31);

			GDALRasterBand *poBand;
			if (xyRaster != NULL)
			{
				// Free Terrain Class on 2nd band (green)
				poBand = newInputDataset->GetRasterBand(2);
				poBand->RasterIO( GF_Write, 0, 0, _xsize, _ysize, 
								  xyRaster, _xsize, _ysize, GDT_Byte, 0, 0 );    
			}

			if (xyRasterBlack != NULL)
			{
				// Rough Terrain Class on 3rd band
				poBand = newInputDataset->GetRasterBand(3);
				poBand->RasterIO( GF_Write, 0, 0, _xsize, _ysize, 
								  xyRasterBlack, _xsize, _ysize, GDT_Byte, 0, 0 );    
				// Other Terrain Class on 1st band
				poBand = newInputDataset->GetRasterBand(1);
				poBand->RasterIO( GF_Write, 0, 0, _xsize, _ysize, 
								  xyRasterBlack, _xsize, _ysize, GDT_Byte, 0, 0 );    
			}

			// We must close the GDAL Dataset in order to the RasterIO writings to be flushed
			//newInputDataset->FlushCache();
			//GDALClose( (GDALDatasetH) newInputDataset);

			// No need to reload the  newInputDataset 
			loadData(dsfilename, newInputDataset);
			// 8 meters used for entropy interpolation
			decompositionLimit = 8.0;
			mapUtility = NULL;
			mapPotential = NULL;
			mapH = NULL;
			initUtility();
		}

		// Freeing unused arrays
		if (xyRaster != NULL)
		{ free(xyRaster); }
		if (xyRasterBlack != NULL)
		{ free(xyRasterBlack); }
	 }/*}}}*/

	 /*------- Constructors --------*/
	 GeoData::~GeoData()
	 { //{{{
		// Delete other objects
		if ( traversabilityBand 	!= NULL )	delete traversabilityBand;
		if ( roughTerrainBand 		!= NULL )	delete roughTerrainBand;
		if ( visibilityBand			!= NULL ) 	delete visibilityBand;
		if ( visibilityHeightBand	!= NULL ) 	delete visibilityHeightBand;
		if ( landmarkBand			!= NULL )	delete landmarkBand; 

		if (dataset != NULL)
		{
			if (dataset->GetRasterCount() < 2) {
				if (mapUtility != NULL)
				{ free(mapUtility); }
				if (mapPotential != NULL)
				{ free(mapPotential); }
				if (mapH != NULL)
				{ free(mapH); }
			}
			if (isInVirtualMemory == 1)
			{
				dataset->FlushCache();
			}
			// GDALClose does not deal correctly with VSI files (in-memory)
			// So only closing datasets from "hard-driven" files
			GDALClose( (GDALDatasetH) dataset);
		}

		if (isInVirtualMemory == 1)
		{
			// First GDAL Dataset must be closed
			VSIUnlink(dsfilename.c_str());
		}
		// Close driver manager : cleaning all registered drivers
		GDALDestroyDriverManager();
	 } //}}}

	 /*------- Other Functions --------*/
	bool GeoData::setUTMProjection(GDALDataset* inputDataset, int utmZone)
	{/*{{{*/
		if ( inputDataset != NULL )
		{
			// Setting GeoSpatial Reference and projection
			OGRSpatialReference oSRS;
			char *pszSRS_WKT = NULL;
			oSRS.SetUTM( utmZone, TRUE );
			oSRS.SetWellKnownGeogCS( "WGS84" );//"EPSG:4326" ); // WGS84
			oSRS.exportToWkt( &pszSRS_WKT );
			inputDataset->SetProjection( pszSRS_WKT );
			CPLFree( pszSRS_WKT );
			return true;
		}
		return false;
	}/*}}}*/

	void GeoData::bufferToMem()
	{//{{{
		if (traversabilityBand != NULL)
		{
			traversabilityBand->bufferToMem();
		}

		if (roughTerrainBand!= NULL)
		{
			roughTerrainBand->bufferToMem();
		}

		if (visibilityBand!= NULL)
		{
			visibilityBand->bufferToMem();
		}

		if (visibilityHeightBand!= NULL)
		{
			visibilityHeightBand->bufferToMem();
		}

	}//}}}

	void GeoData::memToBuffer()
	{//{{{
		if (traversabilityBand != NULL)
		{
			traversabilityBand->memToBuffer();
		}

		if (roughTerrainBand!= NULL)
		{
			roughTerrainBand->memToBuffer();
		}

		if (visibilityBand!= NULL)
		{
			visibilityBand->memToBuffer();
		}

		if (visibilityHeightBand!= NULL)
		{
			visibilityHeightBand->memToBuffer();
		}

	}//}}}

	bool GeoData::setData(int at_i, int at_j, Attributes inatt, bool internalTransfer)
	{//{{{
		int indexPosition  =  (at_j * xsize + at_i);

		setValueInBand(GEODATA_TFLAT_BAND,   indexPosition, inatt.certainty[T_FREE]);
		setValueInBand(GEODATA_TROUGH_BAND,  indexPosition, inatt.certainty[T_ROUGH]);

		if (internalTransfer)
		{
			memToBuffer();
		}

		return true;
	}//}}}

	bool GeoData::setData(double* bufferedData, double* bufferedOther, int _xsize, int _ysize, int _band)
	{//{{{
		if (dataset == NULL || ((bufferedData==NULL) && (bufferedOther==NULL)))
		{
			return false;
		}

		if ((traversabilityBand != NULL) && (bufferedData!=NULL))
		{
			memcpy((traversabilityBand->getValuePtr()), bufferedData, _xsize*_ysize*sizeof(double));
		}

		if ((roughTerrainBand != NULL) && (bufferedOther!=NULL))
		{
			memcpy((roughTerrainBand->getValuePtr()), bufferedOther, _xsize*_ysize*sizeof(double));
		}
		// Writing modification to the GDal dataset 
		memToBuffer();

		return true;
	}//}}}

	void GeoData::setDecompositionLimit(double declim)
	{//{{{
		decompositionLimit = declim;
		if (traversabilityBand != NULL)
		traversabilityBand->setDecompositionLimit(declim);
		if (roughTerrainBand != NULL)
		roughTerrainBand->setDecompositionLimit(declim);
		if (landmarkBand != NULL)
		landmarkBand->setDecompositionLimit(declim);
	}/*}}}*/

	bool GeoData::loadBand(GEODATA_BAND band, const std::string& _dsfilename)
	{ /*{{{*/
		// Open the file and load a GDALDataset
		GDALDataset *ds = (GDALDataset *) GDALOpen( _dsfilename.c_str(), GA_ReadOnly );
		
		if ((ds==NULL)||(ds->GetRasterCount()==0))
		  return false;
		
		// Check size of first BAND
		GDALRasterBand *dsBand = ds->GetRasterBand(1);
		if ((dsBand->GetXSize()!=dataset->GetRasterXSize())
		  || (dsBand->GetYSize()!=dataset->GetRasterYSize()))
		{
			return false;		
		}

		return true;
	} /*}}}*/

	bool GeoData::loadData(const std::string& _dsfilename, GDALDataset* _useThisDataset)
	{ /*{{{*/
		//--- Open and read data 
		// Create the GDAL Dataset
		//dataset = (GDALDataset *) GDALOpen( _dsfilename.c_str(), GA_ReadOnly );
		if ( _useThisDataset == NULL)
		{
			// Opening image source file
			dataset = (GDALDataset *) GDALOpen( _dsfilename.c_str(), GA_Update );
		} else {
			dataset = _useThisDataset;
		}

		// Exit if the GDAL Dataset could not be created
		if( dataset == NULL ) {
			std::cout << "#EEE# [GeoData] Read data file error." << std::endl;
			return false;
		}

		// Initializing affine transformation to default transformation
		refGeoTransform[0] = 0.0;
		refGeoTransform[1] = 1.0; // Width
		refGeoTransform[2] = 0.0;
		refGeoTransform[3] = 0.0;
		refGeoTransform[4] = 0.0;
		refGeoTransform[5] = 1.0; // Height (could be -1.0 in some case)

		std::string projectionRef = std::string( dataset->GetProjectionRef());
		if ( projectionRef.length() == 0 )
		{
			//#ifdef LGL_VERBOSE
			std::cout << "#WWW# [GeoData::loadData] Could get any projection type. Setting default projection." <<  std::endl;
			//#endif
			// This is no projection. This is a simple raster file.
			// We have to set the affine transformation ourself.
			dataset->SetGeoTransform(refGeoTransform);
		}

		// Update the GeoTransformation information
		if((refGeoTransformRType=dataset->GetGeoTransform( (double*)refGeoTransform )) != CE_None)
		{
			//#ifdef LGL_VERBOSE
			std::cerr << "#WWW# [GeoData::loadData] No geo-reference in current dataset" << std::endl;
			//#endif
		}

		// Check the number of bands
		if (dataset->GetRasterCount() >= 2) {
			// Loads the band number GEODATA_TFLAT_BAND which is considered as a traversability information band
			traversabilityBand = new GDInfoBand(dataset,dataset->GetRasterBand(2));
			roughTerrainBand = new GDInfoBand(dataset,dataset->GetRasterBand(3));
			landmarkBand = new GDLandmarkBand(dataset,dataset->GetRasterBand(1));
		} else if (dataset->GetRasterCount()==1) {
			traversabilityBand = new GDInfoBand(dataset, dataset->GetRasterBand(1));
		}

		xsize = dataset->GetRasterXSize();
		ysize = dataset->GetRasterYSize();
		xroot = 0;
		yroot = 0;
		return true;
	} /*}}}*/

	bool GeoData::saveData(const std::string& outfile, const char* wFormat) 
	{ //{{{
		if (dataset == NULL)
		{
			return false;
		}

		// Writing modification to the band
		memToBuffer();

		if (landmarkBand != NULL)
		{
			landmarkBand->memToBuffer();
		}


		//const char *wFormat = savingFormat;
		GDALDriver *wDriver;

		wDriver = GetGDALDriverManager()->GetDriverByName(wFormat);

		if (wDriver == NULL)
		{
			return false;
		}

		GDALDataset* outputDataset = wDriver->CreateCopy( outfile.c_str(), dataset, FALSE, NULL, NULL, NULL );

		// Once we're done, close properly the dataset. This clears everything
		if( outputDataset != NULL )
		{
			GDALClose( (GDALDatasetH) outputDataset);
			return true;
		}

		return false;
	} //}}}

	bool GeoData::saveTraversability(const std::string& outfile)
	{ //{{{
		// SAVING PPM rgb Files 
		// P6
		// # Comments with 6 width and 2 height
		// 6 2
		// 255
		// binary rgb pixels
		if (traversabilityBand == NULL && roughTerrainBand == NULL)
		{
			return false;
		}

		std::ofstream myFile (outfile.c_str(), std::ios::out | std::ios::binary);
		
		if ( ! myFile.good() )
		{
			return false;
		}

		if (roughTerrainBand != NULL)
		{
			char* header= (char*)calloc(256, sizeof(char));
			sprintf(header, "P6\n%d %d\n255\n", getxsize(), getysize());
			myFile << header;
			for (int k=0; k < (getxsize()*getysize()); k++)
			{
				//myFile.put((char)((1.0 - traversabilityBand->getValue(k)) * 255.0));
				myFile.put((char)(( 1.0- (traversabilityBand->getValue(k)+roughTerrainBand->getValue(k) )) * 255.0));
				myFile.put((char)(traversabilityBand->getValue(k) * 255.0));
				myFile.put((char)(roughTerrainBand->getValue(k) * 255.0));
			}
			myFile.close();
			free(header);
		} else {
			char* header= (char*)calloc(256, sizeof(char));
			sprintf(header, "P6\n%d %d\n255\n", getxsize(), getysize());
			myFile << header;
			for (int k=0; k < (getxsize()*getysize()); k++)
			{
				myFile.put((char)((1.0 - traversabilityBand->getValue(k)) * 255.0));
				myFile.put((char)(traversabilityBand->getValue(k) * 255.0));
				myFile.put((char)0);
			}
			myFile.close();
			free(header);

		}
		return true;
	} //}}}

	bool GeoData::saveUtility(const std::string& outfile)
	{ //{{{
		// SAVING PPM greyscale Files 
		// P5
		// # Comments with 6 width and 2 height
		// 6 2
		// 255
		// binary gray pixels

		std::ofstream myFile (outfile.c_str(), std::ios::out | std::ios::binary);
		
		if ( ! myFile.good() )
		{
			return false;
		}

		char* header= (char*)calloc(256, sizeof(char));
		sprintf(header, "P5\n%d %d\n255\n", getxsize(), getysize());

		myFile << header;
		for (int k=0; k < (getxsize()*getysize()); k++)
		{
			myFile.put((char)(mapUtility[k] * 255.0));
			//myFile.put((char)(traversabilityBand->getValue(k) * 255.0));
			//myFile.put((char)(roughTerrainBand->getValue(k) * 255.0));
		}

		myFile.close();
		free(header);
		return true;
	} //}}}

	bool GeoData::saveEntropy(const std::string& outfile)
	{ //{{{
		// SAVING PPM greyscale Files 
		// P5
		// # Comments with 6 width and 2 height
		// 6 2
		// 255
		// binary gray pixels

		std::ofstream myFile (outfile.c_str(), std::ios::out | std::ios::binary);
		
		if ( ! myFile.good() )
		{
			return false;
		}

		char* header= (char*)calloc(256, sizeof(char));
		sprintf(header, "P5\n%d %d\n255\n", getxsize(), getysize());

		myFile << header;
		for (int k=0; k < (getxsize()*getysize()); k++)
		{
			myFile.put((char)(mapH[k] * 255.0));
		}

		myFile.close();
		free(header);
		return true;
	} //}}}

	bool GeoData::savePotential(const std::string& outfile)
	{ //{{{
		// SAVING PPM greyscale Files 
		// P5
		// # Comments with 6 width and 2 height
		// 6 2
		// 255
		// binary gray pixels

		std::ofstream myFile (outfile.c_str(), std::ios::out | std::ios::binary);
		
		if ( ! myFile.good() )
		{
			return false;
		}

		char* header= (char*)calloc(256, sizeof(char));
		sprintf(header, "P5\n%d %d\n255\n", getxsize(), getysize());

		myFile << header;
		for (int k=0; k < (getxsize()*getysize()); k++)
		{
			myFile.put((char)(mapPotential[k] * 255.0));
		}

		myFile.close();
		free(header);
		return true;
	} //}}}

	void GeoData::resetBands() 
	{/*{{{*/
		traversabilityBand->reset(0.5);
	}/*}}}*/

	bool GeoData::createNewInfoBand(GDInfoBand** thatNewBand)
	{//{{{
		// Should never call memToBuffer on this one (same as last band)
		// TODO: its use is still buggy because of GDAL shared objects
		(*thatNewBand) = new GDInfoBand(dataset,dataset->GetRasterBand((dataset->GetRasterCount())-1));
		if ( (*thatNewBand) != NULL)
		{
			return true;
		}
		return false;
		/* GDAL: If the supported format allows it:
		if (dataset != NULL)
		{
			if(dataset->AddBand(GDT_Float64) == CE_None) 
			{
				(*thatNewBand) = new GDInfoBand(dataset,dataset->GetRasterBand((dataset->GetRasterCount())-1));
				return true;
			}
		}
		return false;
		*/
	}//}}}

	//------------------------------------------------------------//

	void GeoData::resetUtility()
	{ //{{{
		int N = getxsize()*getysize();
		for (int k=0; k < N; k++)
		{
			mapPotential[k] = 0.0;
			mapUtility[k] = 0.0;
			mapH[k] = 0.0;
		}
	} //}}}

	bool GeoData::initUtility()
	{//{{{
		int N = getxsize()*getysize();
		if (mapUtility == NULL)
		{
			mapUtility = (double*)calloc(N, sizeof(double));
			mapPotential = (double*)calloc(N, sizeof(double));
			mapH = (double*)calloc(N, sizeof(double));
		} else {
			// in else so it serves as a reset if the maps has already been
			// created
			for (int k=0; k < N; k++)
			{
				mapUtility[k] = 0.0;
				// Zero as init value since the default is "no" potential like zero potential
				mapPotential[k] = 0.0;
				// -1.0 as value to be able to know what to interpolate
				mapH[k] = 0.0;
			}
		}

		return true;
	}//}}}

	bool GeoData::buildUtility()
	{/*{{{*/
		double H=0.0;
		double Hacc=0.0;
		double Umax = 0.0;
		double ndiv = 0.0;
		int Nmax = (getxsize()*getysize());
		int N = 0;
		int Njump = 0;
		int Ndelta = (int)(decompositionLimit/3.0);
		if (roughTerrainBand == NULL)
		{
			for (int k=0; k < Nmax; k++)
			{
				if (Ndelta > 0)
				{
					// -- Averaging entropy on a square sensor foot-print
					Hacc = 0.0;
					ndiv = 0.0;
					Njump = k-Ndelta-(Ndelta*getxsize());
					for ( N = Njump;
						  N <= (k+Ndelta+Ndelta*getxsize()) && N < Nmax;
						  N++)
					{
						if (N < 0)
						{
							N = 0;
							Njump = 0 - Ndelta;
						}
						if (N > (Njump + 2*Ndelta))
						{
							N = Njump + getxsize();
							if(N >= Nmax)
							{
								N = Nmax - 1;
							}
							Njump = N;
						}

						H = 1.0 - (traversabilityBand->getValue(N));
						if (H > 0.0 && H < 1.0)
						{
							H = -H* log(H)- (1.0 -H) * log(1.0 - H);
						} else {
							H = 0.0;
						}
						Hacc += H;
						ndiv = ndiv + 1.0;
					}

					// Weighting Hacc to get final H
					if (ndiv > 0.0)
					{
						H = Hacc / ndiv;
					} else {
						H = 1.0 - (traversabilityBand->getValue(k));
						H = -H* log(H)- (1.0 -H) * log(1.0 - H);
					}

				} else {
					H = 1.0 - (traversabilityBand->getValue(k));
					H = -H* log(H)- (1.0 -H) * log(1.0 - H);
				}

				// Updating entropy only
				mapH[k] = H;

				// Information and Potential fusion
				if (H <= 0.05)
				{
					// Really no information needed on cell k.
					mapUtility[k] = 0.0;
				} else {
					mapUtility[k] = (H+mapPotential[k])/2.0;
				}

				if (mapUtility[k] > Umax)
				{
					Umax = mapUtility[k];
				}
			}
			// Normalizing
			if (Umax > 0.0)
			{
				for (int k=0; k < (getxsize()*getysize()); k++)
				{
					mapUtility[k] = mapUtility[k] / Umax;
					//mapUtility[k] = mapUtility[k] * mapUtility[k] ;
					//mapUtility[k] = mapUtility[k] * mapUtility[k] ;
				}
			}
			return true;
		}

		for (int k=0; k < (getxsize()*getysize()); k++)
		{
			H = 1.0 - (traversabilityBand->getValue(k) + roughTerrainBand->getValue(k));
			if (H == 0.0 || H==1.0)
			{
				H = 0.0;
			} else {
				H = -H* log(H)- (1.0 -H) * log(1.0 - H);
			}
			mapUtility[k] = (H+mapPotential[k])/2.0;
			if (mapUtility[k] > Umax)
			{
				Umax = mapUtility[k];
			}
		}
		// Normalizing
		if (Umax > 0.0)
		{
			for (int k=0; k < (getxsize()*getysize()); k++)
			{
				mapUtility[k] = mapUtility[k] / Umax;
				//mapUtility[k] = mapUtility[k] * mapUtility[k] ;
				//mapUtility[k] = mapUtility[k] * mapUtility[k] ;
			}
		}
		return true;
	}/*}}}*/

	bool GeoData::setMap(GEODATA_INTERNAL_MAP map2set, double pot1_, int x1_, int y1_, double pot2_, int x2_, int y2_)
	{/*{{{*/
		//center concerned cell (x_ y_) is a position manager location 
		double Cx1 = (double)(x1_/2) + 0.5;
		double Cy1 = (double)(y1_/2) + 0.5;
		double Cx2 = (double)(x2_/2) + 0.5;
		double Cy2 = (double)(y2_/2) + 0.5;
		// Two closest points
        double Ix = 0.0; 
        double Iy = 0.0; 
		// Moving coefficient
        double dx = 0.0;
        double dy = 0.0;
		// Indexes
		int i,j;
		// Calculated damping Fact
		double i1 = 0.0;
		double j1 = 0.0;
		double newVal = 0.0;
		double fullDist  = 0.0;
		double dist  = 0.0;
		double dist1 = 0.0;

		// --- Adjusting border indexes
		// Positions are considered to belong to the next pixel
		// but borders can not satisfy this. Thus we constrains
		// borders to the previous pixel (colomns or lines).
		// So we move the cell center accordingly.
		if ((x1_/2) >= xsize)
		{
			Cx1 = Cx1 - 1.0;
		}
		if ((y1_/2) >= ysize)
		{
			Cy1 = Cy1 - 1.0;
		}
		if ((x2_/2) >= xsize)
		{
			Cx2 = Cx2 - 1.0;
		}
		if ((y2_/2) >= ysize)
		{
			Cy2 = Cy2 - 1.0;
		}

		// --- Setting potential on segment ends
		i = (int)Cx1;
		j = (int)Cy1;
		switch(map2set) {
			case ENTROPY_MAP:
				mapH[i + j*xsize] = pot1_;
				break;
			case POTENTIAL_MAP:
				if (mapPotential[i + j*xsize] == 0.0) { mapPotential[i + j*xsize] = pot1_; }
				else { mapPotential[i + j*xsize] = (mapPotential[i + j*xsize] + pot1_)/2.0; }
				break;
			default:
				mapUtility[i + j*xsize] = pot1_;
		}
		i = (int)Cx2;
		j = (int)Cy2;
		switch(map2set) {
			case ENTROPY_MAP:
				mapH[i + j*xsize] = pot2_;
				break;
			case POTENTIAL_MAP:
				if (mapPotential[i + j*xsize] == 0.0) { mapPotential[i + j*xsize] = pot2_; }
				else { mapPotential[i + j*xsize] = (mapPotential[i + j*xsize] + pot2_)/2.0; }
				mapPotential[i + j*xsize] = pot2_;
				break;
			default:
				mapUtility[i + j*xsize] = pot2_;
		}

		// Initialization of distances to the greates distance
		fullDist = euclideanDistance(Cx1, Cy1, Cx2, Cy2);
		if (fullDist <= 0.0)
		{
			return false;
		}

		//--- Spreading potential between the two points
		// Initializing the first search seed
		//Ix = Cx1;
		//Iy = Cy1;
		// Look around for the two closest to (Cx2,Cy2)
		// (Cx1, Cy1) is moving towards (Cx2,Cy2)
		while(!(Cx1 == Cx2 && Cy1 == Cy2))
		{
			dist1 = fullDist;
			i1 = -1.0;
			j1 = -1.0;
			// Find the next closest neighbors to Cx2 Cy2
			// To draw the line between point 1 and point 2
			for (dx = -1.0; dx <= 1.0; dx = dx + 1.0)
			{
				for (dy = -1.0; dy <= 1.0; dy = dy + 1.0)
				{
					Ix = Cx1 + dx;
					Iy = Cy1 + dy;
					if (!(dy == 0.0 && dx == 0.0))
					{
						dist = euclideanDistance(Ix, Iy, Cx2, Cy2);
						if (dist1 > dist)
						{
							dist1 = dist;
							i1 = Ix;
							j1 = Iy;
						}
					}
				} // eof
			} // eof

			// Setting next search start-point as this closest
			Cx1 = i1;
			Cy1 = j1;
			if (!(Cx1 == Cx2 && Cy1 == Cy2))
			{
				// Set the potential on i1,j1 and i2,j2
				i = (int)i1;
				j = (int)j1;
				if (i >= 0 && j >=0 && i < xsize && j < ysize)
				{
					newVal = (dist1 / fullDist); // weight for pot1_
					newVal = newVal * pot1_ + (1.0 - newVal) * pot2_;
					switch(map2set) {
						case ENTROPY_MAP:
							mapH[i + j*xsize] = newVal;
							break;
						case POTENTIAL_MAP:
							//mapPotential[i + j*xsize] = newVal;
							if (mapPotential[i + j*xsize] == 0.0) { mapPotential[i + j*xsize] = newVal; }
							else { mapPotential[i + j*xsize] = (mapPotential[i + j*xsize] + newVal)/2.0; }
							break;
						default:
							mapUtility[i + j*xsize] = newVal;
					}
				}
			}
		} // eow

		return true;
	}/*}}}*/

	double GeoData::euclideanDistance(double ax, double ay, double bx, double by)
	{/*{{{*/
		return sqrt((ax-bx)*(ax-bx) + (ay-by)*(ay-by));
	}/*}}}*/

	bool GeoData::loadInfoBand(const std::string& datafile, GEODATA_BAND aband)
	{ //{{{

		if (datafile.empty())
		  return false;
		
		/* Open and read data */
		GDALDataset *loadedDataset = (GDALDataset *) GDALOpen( datafile.c_str(), GA_ReadOnly );
		if( loadedDataset == NULL ) {
		  std::cout << "#EEE# [Geodata] Read data file error"<< std::endl;	 
		}
	 
		/* check the number of bands */
		if (loadedDataset->GetRasterCount()!=1)
		  return false;
  
		/* copy band */
		// if a trav band exists
		switch(aband)
		{
			case GEODATA_TFLAT_BAND:
				if (traversabilityBand != NULL) {
					// Fill it
				} else {
					// Create it
				}
				break;
			default:
				break;
		}
  
  		// Cleaning unused data
		GDALClose( (GDALDatasetH) loadedDataset);
		return true;
	} //}}}

	bool GeoData::loadLandmarkBand(const std::string& _dsfilename)
	 { //{{{
		return true;
	 } //}}}

	bool GeoData::isVisionTransparent(int atx, int aty, int atz)
	{//{{{
		// Sum of probabilities on traversable land
		// simulationProbaVisionLimit
		double psumOfTTerrain = 0.0;
		if (traversabilityBand != NULL)
		{ psumOfTTerrain = traversabilityBand->getValue(atx + xsize*aty); }

		if (roughTerrainBand != NULL)
		{ psumOfTTerrain = psumOfTTerrain + roughTerrainBand->getValue(atx + xsize*aty); }

		if (psumOfTTerrain > simulationProbaVisionLimit )
		{
			// This terrain is somehow considered traversable
			// thus considered as we can see through it !
			return true;
		}

		return false;

	}//}}}

	//-------------------------- GETTERS ----------------------------//

	 GDALDataset* GeoData::getDataset() const {//{{{
		return dataset;
	 }//}}}

	void GeoData::getAffineTransform ( double& utmXRoot, double& utmYRoot, double& cellXRes, double& cellYRes) {/*{{{*/
		utmXRoot = refGeoTransform[0];
		utmYRoot = refGeoTransform[3];
		cellXRes = refGeoTransform[1];
		cellYRes = refGeoTransform[5];
	}/*}}}*/

	 double GeoData::geoxroot() const {/*{{{*/
		return refGeoTransform[0];
	 }/*}}}*/

	 double GeoData::geoyroot() const {/*{{{*/
		return refGeoTransform[3];
	 }/*}}}*/

	 double GeoData::geoxend() const {/*{{{*/
		return (geoxroot() + (getxsize() * refGeoTransform[1]));
	 }/*}}}*/

	 double GeoData::geoyend() const {/*{{{*/
		return (geoyroot() + (getysize() * refGeoTransform[5]));
	 }/*}}}*/

	 double GeoData::geoxres() const {/*{{{*/
		return refGeoTransform[1];
	 }/*}}}*/

	 double GeoData::geoyres() const {/*{{{*/
		return refGeoTransform[5];
	 }/*}}}*/

	 //--- Abstract methods from abstract class Raster 
	 int GeoData::getxroot() const {/*{{{*/
		return 0;
	 }/*}}}*/
	
	 int GeoData::getyroot() const {//{{{
		return 0;
	 }//}}}

	 int GeoData::getxsize() const { //{{{
		return xsize; //dataset->GetRasterXSize();
	 } //}}}

	 int GeoData::getysize() const { //{{{
		return ysize; //dataset->GetRasterYSize();
	 }//}}}
	
	//--- Getting data from a special BAND
	double GeoData::getValueFromBand(GEODATA_BAND _aband, int xindex, int yindex)
	{//{{{
		if ( xindex <  this->getxroot() ||
			 xindex >= this->getxsize() ||
			 yindex <  this->getyroot() ||
			 yindex >= this->getysize() )
			 	return ( -1.0 ) ;
		
		int index = xindex + yindex* xsize;
		return ( getValueFromBand(_aband, index) ) ;

	}//}}}

	double GeoData::getValueFromBand(GEODATA_BAND _aband, int index)
	{//{{{
		// It is very important to get Zero values on non-existing bands if at
		// least the traversabilityBand is present.
		// If not present then it means no map connexion has been established
		// and the default return -1.0 is desired.
		switch(_aband)
		{
			case GEODATA_TFLAT_BAND:
				if (traversabilityBand != NULL)
				{ return traversabilityBand->getValue(index); }
				break;
			case GEODATA_TROUGH_BAND:
				if (roughTerrainBand != NULL)
				{ return roughTerrainBand->getValue(index); }
				if (roughTerrainBand == NULL && traversabilityBand != NULL)
				{ return 0.0; }
				break;
			case GEODATA_LANDMARK_BAND:
				if (landmarkBand != NULL)
				{ return landmarkBand->getValue(index); }
				if (landmarkBand == NULL && traversabilityBand != NULL)
				{ return 0.0; }
				break;
			case GEODATA_VISIBILITY_BAND:
				if (visibilityBand != NULL)
				{ return visibilityBand->getValue(index); }
				break;
			case GEODATA_VISIBILITY_HEIGHT_BAND:
				if (visibilityHeightBand != NULL)
				{ return visibilityHeightBand->getValue(index); }
				break;
			default:
				return -1.0;
		}
		return -1.0;
	}//}}}

	void GeoData::setValueInBand(GEODATA_BAND _aband, int _i, double _val)
	{//{{{
		switch(_aband)
		{
			case GEODATA_TROUGH_BAND:
				if (roughTerrainBand != NULL)
				{ roughTerrainBand->setValue(_i, _val); }
				break;
			case GEODATA_LANDMARK_BAND:
				if (landmarkBand != NULL)
				{ landmarkBand->setValue(_i, _val); }
				break;
			case GEODATA_TFLAT_BAND:
				if (traversabilityBand != NULL)
				{ traversabilityBand->setValue(_i, _val); }
				break;
			default:
				if (traversabilityBand != NULL)
				{
					traversabilityBand->setValue(_i, _val);
				}
		}
	}//}}}

	 //--- get geo data value for the raster cell (x,y)
	 double GeoData::getData(const RasterCellIndex& cell) const
	 { //{{{
		double result=0.0;
		double w=0.0;
		
		// the block size (preferred block size for each reading) must be 1
		if (traversabilityBand!=NULL) {
			w = w + traversabilityBand->getWeight();
		  result+=traversabilityBand->getWeight()*traversabilityBand->get(cell); 
		} else { std::cout << "#EEE# [GeoData] Error: Traversability Band is NULL" << std::endl; }
		
		if (roughTerrainBand !=NULL) {
			w = w + roughTerrainBand->getWeight();
			result+=roughTerrainBand->getWeight()*roughTerrainBand->get(cell); 
		}
		/*
		bool onlyTrav=false;
		if (landmarkBand!=NULL) {
		  result+= landmarkBand->getWeight()*landmarkBand->get(cell);
		} else {
		  //std::cout << "[GeoData] Error: Landmark Band is NULL" << std::endl;
		  onlyTrav=true;
		}
		*/

		//result /= (onlyTrav==true)?1.0:(roughTerrainBand->getWeight()+landmarkBand->getWeight()+traversabilityBand->getWeight());
		
		// do not do (result/w)  since P(trav+roughTrav) = 1 - P(obstacle)
		return result;
	 }//}}}

	 double GeoData::getData(GEODATA_BAND band, const RasterCellIndex& cell) const
	 { //{{{
		double value=0.0;
		// the block size (preferred block size for each reading) must be 1
		if (band == GEODATA_TFLAT_BAND)
		  value=traversabilityBand->get(cell);
		else if (band == GEODATA_LANDMARK_BAND)
		  value=landmarkBand->get(cell);
		else if (band == GEODATA_TROUGH_BAND)
		  value=roughTerrainBand->get(cell);
		else if (band == GEODATA_VISIBILITY_BAND)
		  value=visibilityBand->get(cell);
		else if (band == GEODATA_VISIBILITY_HEIGHT_BAND)
		  value=visibilityHeightBand->get(cell);
		return value;
	 } //}}}

	 double GeoData::getData(const RasterRect& rect) const
	 { //{{{
		double result=0.0;
		// TODO: homogeneize this with NavGraph costs (see NavGraph::navCostAtNode)
		// Arbitrary costs for (in order):
		// traversable terrain - rough terrain - obstacles
		//double costw[3] = {1.0/1.0,  1.0/3.0, 1.0/6.0};

		int i=0;
		std::vector<double> travValues;
		//std::vector<double> landmarkValues;
		std::vector<double> roughtravValues;
		//int travWeight = 0;
		//int landmarkWeight = 0;
		
		if (rect.xsize()*rect.ysize() <= 0.0)
		{
			return -1.0;
		}

		/* test if trav and landmark band are present */
		if ((traversabilityBand==NULL) && (landmarkBand==NULL) && (roughTerrainBand==NULL)) {
		  std::cout << "#EEE# [GeoData] get data error: No bands present." << std::endl;
		  return 0.0;
		}
		
		if (traversabilityBand != NULL)
		{
			if (!traversabilityBand->get(rect,travValues)) {
			  std::cout << "#EEE# [GeoData] Could not access traversability band data" << std::endl;		  
			  return -1.0;
			}
		}

		if (roughTerrainBand != NULL)
		{
			if (!roughTerrainBand->get(rect,roughtravValues)) {
			  std::cout << "#EEE# [GeoData] Could not access rough terrain band data" << std::endl;		  
			  return -1.0;
			}
		}
		
		if ((traversabilityBand != NULL) && (roughTerrainBand != NULL))
		{
			for (i=0;i<rect.xsize()*rect.ysize();i++) {
				result += ( travValues[i] + /*0.333 == costw[1]*/ 0.3333333333*roughtravValues[i] );
			}
		} else {
			if (traversabilityBand != NULL)
			{
				for (i=0;i<rect.xsize()*rect.ysize();i++) {
					result += travValues[i];
				}
			} else {
				result = 0.0;
			}
		}

		// Average making
		result /= (rect.xsize()*rect.ysize());

		return result;
	 } //}}}

	 double GeoData::getData(GEODATA_BAND band, const RasterRect& rect) const 
	 { //{{{
		double result=0.0;
		
		return result;
	 } //}}}

	double GeoData::getScale() const 
	{ /*{{{*/
		if (dataset == NULL) {  
			return GEODATA_DEFAULT_SCALE();
		}
		if (dataset->GetRasterCount()<=0)
		{
			return GEODATA_DEFAULT_SCALE();
		} else {
			/* We use the scale of the first band.
			 * we suppose that all band scales have the same value
			 */
			return (double)(dataset->GetRasterBand(0+1)->GetScale());
		}
	} /*}}}*/

	 GDRasterBand<double>* GeoData::getBand(GEODATA_BAND band)
	 { //{{{
		if (band == GEODATA_TFLAT_BAND)
		  return traversabilityBand;
		else if (band == GEODATA_LANDMARK_BAND)
		  return landmarkBand;
		else if (band == GEODATA_TROUGH_BAND)
		  return roughTerrainBand;
		return NULL;
	 } //}}}

	double GeoData::getInterpolatedValue(GEODATA_BAND _aband, double xNavLoc, double yNavLoc)
	{ //{{{
		// For other bands than Obstacles
		GDRasterBand<double>* thatBand =  NULL;
		if (_aband == GEODATA_TRICLASSOBST_BAND)
		{
			thatBand = getBand(GEODATA_TFLAT_BAND);
			GDRasterBand<double>* otherBand = getBand(GEODATA_TROUGH_BAND);
			// Interpolate the position
			// navLoc positions are grid-doubled 
			xNavLoc = xNavLoc /2.0;
			yNavLoc = yNavLoc /2.0;

			if (   (int)xNavLoc == getxsize() || (int)yNavLoc == getysize() 
			    || (int)xNavLoc == getxroot() || (int)yNavLoc == getyroot() )
			{
				// catching all border cases
				if ((int)xNavLoc == getxsize()) 
				{ xNavLoc = xNavLoc - 1.0; }
				if ((int)yNavLoc == getysize()) 
				{ yNavLoc = yNavLoc - 1.0; }
			} else {
				// Off border effects
				if ((double)((int)(xNavLoc/2.0))*2.0 == xNavLoc)
				{
					if ((double)((int)(yNavLoc/2.0))*2.0 == yNavLoc)
					{
						// We are on a corner
						double tmp1 = thatBand->getValue(RasterPoint((int)xNavLoc, (int)yNavLoc));
						double tmp2 = thatBand->getValue(RasterPoint(((int)xNavLoc)-1, (int)yNavLoc));
						double tmp3 = thatBand->getValue(RasterPoint((int)xNavLoc, ((int)yNavLoc)-1));
						double tmp4 = thatBand->getValue(RasterPoint(((int)xNavLoc)-1, ((int)yNavLoc)-1));
						// Complementary values
						double tmpc1 = 0.0;
						double tmpc2 = 0.0;
						double tmpc3 = 0.0;
						double tmpc4 = 0.0;
						if (otherBand != NULL)
						{
							tmpc1 = otherBand->getValue(RasterPoint((int)xNavLoc, (int)yNavLoc));
							tmpc2 = otherBand->getValue(RasterPoint(((int)xNavLoc)-1, (int)yNavLoc));
							tmpc3 = otherBand->getValue(RasterPoint((int)xNavLoc, ((int)yNavLoc)-1));
							tmpc4 = otherBand->getValue(RasterPoint(((int)xNavLoc)-1, ((int)yNavLoc)-1));
						}
						tmp1 = 1.0 - (tmp1 + tmpc1);
						tmp2 = 1.0 - (tmp2 + tmpc2);
						tmp3 = 1.0 - (tmp3 + tmpc3);
						tmp4 = 1.0 - (tmp4 + tmpc4);
						return ((tmp1+tmp2+tmp3+tmp4)/4.0);
					} else {
						// We are between two colmuns : acting X
						double tmp1 = thatBand->getValue(RasterPoint((int)xNavLoc, (int)yNavLoc));
						double tmp2 = thatBand->getValue(RasterPoint(((int)xNavLoc)-1, (int)yNavLoc));
						// Complementary values
						double tmpc1 = 0.0;
						double tmpc2 = 0.0;
						if (otherBand != NULL)
						{
							tmpc1 = otherBand->getValue(RasterPoint((int)xNavLoc, (int)yNavLoc));
							tmpc2 = otherBand->getValue(RasterPoint(((int)xNavLoc)-1, (int)yNavLoc));
						}
						tmp1 = 1.0 - (tmp1 + tmpc1);
						tmp2 = 1.0 - (tmp2 + tmpc2);
						return ((tmp1+tmp2)/2.0);
					}
				} else {
					// We are between two rows : acting Y
					if ((double)((int)(yNavLoc/2.0))*2.0 == yNavLoc)
					{
						// We are between two lines : acting on Y
						double tmp1 = thatBand->getValue(RasterPoint((int)xNavLoc, (int)yNavLoc));
						double tmp2 = thatBand->getValue(RasterPoint((int)xNavLoc, ((int)yNavLoc)-1));
						// Complementary values
						double tmpc1 = 0.0;
						double tmpc2 = 0.0;
						if (otherBand != NULL)
						{
							tmpc1 = otherBand->getValue(RasterPoint((int)xNavLoc, (int)yNavLoc));
							tmpc2 = otherBand->getValue(RasterPoint((int)xNavLoc, ((int)yNavLoc)-1));
						}
						tmp1 = 1.0 - (tmp1 + tmpc1);
						tmp2 = 1.0 - (tmp2 + tmpc2);
						return ((tmp1+tmp2)/2.0);
					} 
				}
			}
			// We are in the middle of a cell
			double tmp  = thatBand->getValue(RasterPoint((int)xNavLoc, (int)yNavLoc));
			double tmpc = otherBand->getValue(RasterPoint((int)xNavLoc, (int)yNavLoc));
			return (1.0-(tmp+tmpc));
		}

		// For other bands than Obstacles
		thatBand =  getBand(_aband);
		if (thatBand == NULL)
		{
			return -1.0;
		} else {
			// Interpolate the position
			// navLoc positions are grid-doubled 
			xNavLoc = xNavLoc /2.0;
			yNavLoc = yNavLoc /2.0;

			if (   (int)xNavLoc == getxsize() || (int)yNavLoc == getysize() 
			    || (int)xNavLoc == getxroot() || (int)yNavLoc == getyroot() )
			{
				// catching all border cases
				if ((int)xNavLoc == getxsize()) 
				{ xNavLoc = xNavLoc - 1.0; }
				if ((int)yNavLoc == getysize()) 
				{ yNavLoc = yNavLoc - 1.0; }
			} else {
				double tmp = 0.0;
				// Off border effects
				if ((double)((int)(xNavLoc/2.0))*2.0 == xNavLoc)
				{
					if ((double)((int)(yNavLoc/2.0))*2.0 == yNavLoc)
					{
						// We are on a corner
						tmp = thatBand->getValue(RasterPoint((int)xNavLoc, (int)yNavLoc));
						tmp = tmp + thatBand->getValue(RasterPoint(((int)xNavLoc)-1, (int)yNavLoc));
						tmp = tmp + thatBand->getValue(RasterPoint((int)xNavLoc, ((int)yNavLoc)-1));
						tmp = tmp + thatBand->getValue(RasterPoint(((int)xNavLoc)-1, ((int)yNavLoc)-1));
						return (tmp/4.0);
					} else {
						// We are between two colmuns : acting X
						tmp = thatBand->getValue(RasterPoint((int)xNavLoc, (int)yNavLoc));
						tmp = tmp + thatBand->getValue(RasterPoint(((int)xNavLoc)-1, (int)yNavLoc));
						return (tmp/2.0);
					}
				} else {
					if ((double)((int)(yNavLoc/2.0))*2.0 == yNavLoc)
					{
						// We are between two lines : acting on Y
						tmp = thatBand->getValue(RasterPoint((int)xNavLoc, (int)yNavLoc));
						tmp = tmp + thatBand->getValue(RasterPoint((int)xNavLoc, ((int)yNavLoc)-1));
						return (tmp/2.0);
					} 
				}
			}
		}

		// We are in the middle of a cell
		return thatBand->getValue(RasterPoint((int)xNavLoc, (int)yNavLoc));

	} //}}}

	double GeoData::getFootPrintEntropy(int xcenter, int ycenter)
	{/*{{{*/
		double H=0.0;
		double P1=0.0;
		double P2=0.0;
		double Hacc=0.0;
		double ndiv = 0.0;
		int Nmax = (getxsize()*getysize());
		int N = 0;
		int Njump = 0;
		int Ndelta = (int)(decompositionLimit/3.0);// sensor foot print like
		int k = xcenter + ycenter * xsize;
		if (Ndelta > 0)
		{
			// -- Averaging entropy on a square sensor foot-print
			Hacc = 0.0;
			ndiv = 0.0;
			Njump = k-Ndelta-(Ndelta*getxsize());
			for ( N = Njump;
				  N <= (k+Ndelta+Ndelta*getxsize()) && N < Nmax;
				  N++)
			{
				if (N < 0)
				{
					N = 0;
					Njump = 0 - Ndelta;
				}
				if (N > (Njump + 2*Ndelta))
				{
					N = Njump + getxsize();
					if(N >= Nmax)
					{
						N = Nmax - 1;
					}
					Njump = N;
				}

				P1 = getValueFromBand(GEODATA_TFLAT_BAND, N);
				P2 = getValueFromBand(GEODATA_TROUGH_BAND, N);
				if    ((P1 > 0.0 && P1 < 1.0)
				    && (P2 > 0.0 && P2 < 1.0))
				{
					H = -P1* log(P1) - P2*log(P2) - (1.0 - (P1+P2)) * log(1.0 - (P1+P2));
				} else {
					if ((P1 > 0.0 && P1 < 1.0)) {
						H = -P1* log(P1) - (1.0 - P1) * log(1.0 - P1);
					} else {
						H = 0.0;
					}
				}
				Hacc += H;
				ndiv = ndiv + 1.0;
			}

			// Weighting Hacc to get final H
			if (ndiv > 0.0)
			{
				H = Hacc / ndiv;
			} else {
				H = getValueFromBand(GEODATA_TFLAT_BAND, N);
				H = -H* log(H)- (1.0 -H) * log(1.0 - H);
			}

		} else {
			H = getValueFromBand(GEODATA_TFLAT_BAND, N);
			H = -H* log(H)- (1.0 -H) * log(1.0 - H);
		}
		return H;
	}/*}}}*/

	HOMOGENEOUS_TYPE GeoData::getTraversabilityHomogeneousType()
	{//{{{
		if(traversabilityBand==NULL) { return HT_NONE; }
		return traversabilityBand->getHomogeneousType();
	}//}}}

	HOMOGENEOUS_TYPE GeoData::getRoughHomogeneousType()
	{//{{{
		if(roughTerrainBand==NULL) { return HT_NONE; }
		return roughTerrainBand->getHomogeneousType();
	}//}}}

	double GeoData::getTraversabilityProbabilityLag()
	{//{{{
		if(traversabilityBand==NULL) { return -1.0; }
		return traversabilityBand->getProbabilityLag();
	}//}}}

	double GeoData::getRoughProbabilityLag()
	{//{{{
		if(roughTerrainBand==NULL) { return -1.0; }
		return roughTerrainBand->getProbabilityLag();
	}//}}}

	double GeoData::getTraversabilitySigmaFactor()
	{//{{{
		if(roughTerrainBand==NULL) { return -1.0; }
		return roughTerrainBand->getSigmaFactor();
	}//}}}

	double GeoData::getRoughSigmaFactor()
	{//{{{
		if(roughTerrainBand==NULL) { return -1.0; }
		return roughTerrainBand->getSigmaFactor();
	}//}}}

	double GeoData::getSimulationProbaVisionLimit()
	{//{{{
		return simulationProbaVisionLimit;
	}//}}}

	//------------------------- SETTERS ------------------------------//

	 void GeoData::setScale(double _scale)
	 { //{{{
		if (dataset->GetRasterCount()<=0)
		  return;
		else {
		  // set the scale of the first band
		  // suppose that all band scales have the same value
		  dataset->GetRasterBand(0)->SetScale((double)_scale);
		}  
	 } //}}}

	void GeoData::setDecompositionType(HOMOGENEOUS_TYPE htype_)
	{//{{{
		if (traversabilityBand != NULL)
		{ traversabilityBand->setHomogeneousType(htype_); }
		if (roughTerrainBand   != NULL)
		{ roughTerrainBand->setHomogeneousType(htype_); }
		if (landmarkBand != NULL)
		{ landmarkBand->setHomogeneousType(htype_); }
	}//}}}

	void GeoData::setProbabilityLag(double probaLag_)
	{//{{{
		if (traversabilityBand != NULL)
		{ traversabilityBand->setProbabilityLag(probaLag_); }
		if (roughTerrainBand   != NULL)
		{ roughTerrainBand->setProbabilityLag(probaLag_); }
		if (landmarkBand != NULL)
		{ landmarkBand->setProbabilityLag(probaLag_); }
	}//}}}

	void GeoData::setSigmaFactor(double sigmaF_)
	{//{{{
		if (traversabilityBand != NULL)
		{ traversabilityBand->setSigmaFactor(sigmaF_); }
		if (roughTerrainBand   != NULL)
		{ roughTerrainBand->setSigmaFactor(sigmaF_); }
		if (landmarkBand != NULL)
		{ landmarkBand->setSigmaFactor(sigmaF_); }
	}//}}}

	void GeoData::setFusingMethod(FUSION_METHOD _fmethod)
	 {/*{{{*/
	 	fmethod = _fmethod;
	 }/*}}}*/

	void GeoData::setSimulationProbaVisionLimit(double spvl_)
	{//{{{
		simulationProbaVisionLimit = spvl_;
	}//}}}

	//---------------------------------------------------------------//

	 // given a rect ((top,left),(right,bottom)), tell if this is homogeneous
	 bool GeoData::isHomogeneous(const RasterRect& rect) const 
	 { //{{{
		if ((traversabilityBand != NULL)&&(landmarkBand == NULL)&&(roughTerrainBand == NULL)) 
		{ return traversabilityBand->isHomogeneous(rect); }

		if  ((traversabilityBand == NULL)&&(landmarkBand != NULL)&&(roughTerrainBand == NULL))
		{ return landmarkBand->isHomogeneous(rect); }

		if  ((traversabilityBand == NULL)&&(landmarkBand == NULL)&&(roughTerrainBand != NULL))
		{ return roughTerrainBand->isHomogeneous(rect); }

		if ((traversabilityBand != NULL)&&(landmarkBand != NULL)&&(roughTerrainBand != NULL))
		{ return (  traversabilityBand->isHomogeneous(rect) 
		         && landmarkBand->isHomogeneous(rect)
				 && roughTerrainBand->isHomogeneous(rect)
				 ); }

		return false;
	 } //}}}

	 bool GeoData::sameValue(const RasterRect& rect) 
	 { //{{{
		bool result=false;
		if (traversabilityBand != NULL)
		  result = traversabilityBand->sameValue(rect);
		if (landmarkBand != NULL)
		  result = result&&landmarkBand->sameValue(rect);
		if (roughTerrainBand != NULL)
		  result = result&&roughTerrainBand->sameValue(rect);
		
		return result;
	 } //}}}

	bool GeoData::sameValue(GEODATA_BAND band, const RasterRect& rect)
	{ //{{{
		// the block size (preferred block size for each reading) must be 1
		if (band == GEODATA_TFLAT_BAND)
			return traversabilityBand->sameValue(rect);
		else if (band == GEODATA_TROUGH_BAND)
			return roughTerrainBand->sameValue(rect);
		else if (band == GEODATA_LANDMARK_BAND)
			return landmarkBand->sameValue(rect);
		return false;
	} //}}}

	void GeoData::addDataset(GDALDataset *_dataset)
	{ //{{{
  
		int iBands;
		GDALRasterBand *band;
		GDALRasterBand *destBand;
  
		if (dataset==NULL) {
			std::cout << "#WWW# GeoData::addDataset: dataset NULL" << std::endl;
			return;
		}
  
		//std::cout << _dataset->GetRasterCount() << std::endl;
  
		for(iBands = 1; iBands <= _dataset->GetRasterCount(); iBands++ )
		{
		  band = _dataset->GetRasterBand(iBands);
	 
		  // add a new band
		  dataset->AddBand(band->GetRasterDataType(),NULL);
		  // Check 
		  if (dataset->GetRasterCount() <= 0)
			 return;
		  destBand = dataset->GetRasterBand(dataset->GetRasterCount());
		}
	
	} //}}}

	bool GeoData::mergeStrict(GeoData* _toMerge, GEODATA_BAND _aband)
	{ //{{{
		//--- Checking overlapping
		if ( !overlaps(_toMerge) )
		{
			return false;
		}

		//--- Getting the common rects
		RasterRect inRect;
		RasterRect fusRect;
		double tmp = 0.0;
		// Getting geographical coordinates bounds on these overlapping geodatas
		// TODO : deal with UTM zone coordinates switches
		
		// XStart
		if ( geoxroot() > _toMerge->geoxroot() )
		{
			// Merging left side is outside
			inRect.setXTopLeft(0);
			if (_toMerge->geoxres() > 0.0)
			{
				// Setting X bound on fusing data
				tmp = geoxroot() - _toMerge->geoxroot();
				tmp = tmp / _toMerge->geoxres();
				if ((tmp - (double)((int)tmp)) > 0.5 )
				{
					fusRect.setXTopLeft(((int)tmp) + 1);
				} else {
					fusRect.setXTopLeft((int)tmp);
				}
			} else { return false;}
		} else {
			// Merging left side is inside
			fusRect.setXTopLeft(0);
			if (geoxres() > 0.0)
			{
				// Setting X bound on fusing data
				tmp = _toMerge->geoxroot() - geoxroot();
				tmp = tmp / geoxres();
				if ((tmp - (double)((int)tmp)) > 0.5 )
				{
					inRect.setXTopLeft(((int)tmp) + 1);
				} else {
					inRect.setXTopLeft((int)tmp);
				}
			} else { return false;}
		}

		// XEnd
		if (geoxend() > _toMerge->geoxend())
		{
			// Merging right side is inside
			// (-1) stands here because we indicate the coordinates and the top left corner of the bottom right pixel
			fusRect.setXBottomRight(_toMerge->getxsize() - 1 );
			if (geoxres() > 0.0)
			{
				// Setting X bound on fusing data
				tmp = geoxend() - _toMerge->geoxend();
				tmp = tmp / geoxres(); // nb of pixels
				if ((tmp - (double)((int)tmp)) > 0.5 )
				{
					inRect.setXBottomRight(getxsize() - ((int)tmp));
				} else {
					inRect.setXBottomRight(getxsize() - (((int)tmp) + 1));
				}
			} else { return false;}

		} else {
			// Merging right side is outside
			inRect.setXBottomRight(getxsize() - 1 );
			if (_toMerge->geoxres() > 0.0)
			{
				// Setting X bound on fusing data
				tmp = _toMerge->geoxend() - geoxend();
				tmp = tmp / _toMerge->geoxres();
				if ((tmp - (double)((int)tmp)) > 0.5 )
				{
					fusRect.setXBottomRight(_toMerge->getxsize() - ((int)tmp));
				} else {
					fusRect.setXBottomRight(_toMerge->getxsize() - (((int)tmp) + 1));
				}
			} else { return false;}

		}

		// For Y Coordinates North is UP so the logic is inverted: but it depends on the geoyres() sign.
		// YStart
		tmp = fabs(_toMerge->geoyroot() - geoyroot()); 
		if ( (geoyres() > 0.0) && (_toMerge->geoyres() > 0.0))// geoyroot() <= _toMerge->geoyroot() )
		{
			if (geoyroot() > _toMerge->geoyroot())
			{
				// Merging top side is outside
				inRect.setYTopLeft(0);
				tmp = tmp / _toMerge->geoyres();
				if ((tmp - (double)((int)tmp)) > 0.5 )
				{
					fusRect.setYTopLeft(((int)tmp) + 1);
				} else {
					fusRect.setYTopLeft((int)tmp);
				}
			} else { 
				// Merging top side is inside
				fusRect.setYTopLeft(0);
				tmp = tmp / geoyres();
				if ((tmp - (double)((int)tmp)) > 0.5 )
				{
					inRect.setYTopLeft(((int)tmp) + 1);
				} else {
					inRect.setYTopLeft((int)tmp);
				}
			}
		} else {
			if ( (geoyres() < 0.0) && (_toMerge->geoyres() < 0.0))
			{
				if (geoyroot() <= _toMerge->geoyroot())
				{
					inRect.setYTopLeft(0);
					tmp = tmp / (0.0 - _toMerge->geoyres());
					if ((tmp - (double)((int)tmp)) > 0.5 )
					{
						fusRect.setYTopLeft(((int)tmp) + 1);
					} else {
						fusRect.setYTopLeft((int)tmp);
					}
				} else { 
					fusRect.setYTopLeft(0);
					tmp = tmp / (0.0-geoyres());
					if ((tmp - (double)((int)tmp)) > 0.5 )
					{
						inRect.setYTopLeft(((int)tmp) + 1);
					} else {
						inRect.setYTopLeft((int)tmp);
					}
				}
			} else { std::cout << "#EEE# Can not merge GeoData with difference resolution signature"<< std::endl; return false;}
		}

		// YEnd
		tmp = fabs(geoyend() - _toMerge->geoyend());
		if ( (geoyres() > 0.0) && (_toMerge->geoyres() > 0.0))// geoyroot() <= _toMerge->geoyroot() )
		{
			if (geoyend() > _toMerge->geoyend())
			{
				// Inside
				fusRect.setYBottomRight(_toMerge->getysize() - 1);
				tmp = tmp / geoyres(); // nb of piyels
				if ((tmp - (double)((int)tmp)) <= 0.5 )
				{
					inRect.setYBottomRight(getysize() - 1 - ((int)tmp));
				} else {
					inRect.setYBottomRight(getysize() - 1 - (((int)tmp) + 1));
				}
			} else { 
				// Outside 
				inRect.setYBottomRight(getysize() - 1);
				tmp = tmp / _toMerge->geoyres(); // nb of piyels
				if ((tmp - (double)((int)tmp)) <= 0.5 )
				{
					fusRect.setYBottomRight(_toMerge->getysize() - 1 - ((int)tmp));
				} else {
					fusRect.setYBottomRight(_toMerge->getysize() - 1 - (((int)tmp) + 1));
				}
			}
		} else {
			if ( (geoyres() < 0.0) && (_toMerge->geoyres() < 0.0))
			{
				if (geoyend() < _toMerge->geoyend())
				{
					// Inside
					fusRect.setYBottomRight(_toMerge->getysize() - 1);
					tmp = tmp / (0.0- geoyres()); // nb of piyels
					if ((tmp - (double)((int)tmp)) <= 0.5 )
					{
						inRect.setYBottomRight(getysize() - 1 - ((int)tmp));
					} else {
						inRect.setYBottomRight(getysize() - 1 - (((int)tmp) + 1));
					}
				} else { 
					// Outside 
					inRect.setYBottomRight(getysize() - 1);
					tmp = tmp / (0.0 - _toMerge->geoyres()); // nb of piyels
					if ((tmp - (double)((int)tmp)) <= 0.5 )
					{
						fusRect.setYBottomRight(_toMerge->getysize() - 1 - ((int)tmp));
					} else {
						fusRect.setYBottomRight(_toMerge->getysize() - 1 - (((int)tmp) + 1));
					}
				}
			} else { std::cout << "#EEE# Can not merge GeoData with difference resolution signature"<< std::endl; return false;}
		}

		//--- Fusing
		fuseData(_toMerge, _aband, inRect, fusRect);
		return true;
	} //}}}

	bool GeoData::overlaps(GeoData* _toMerge)
	{ //{{{
		if (   geoxroot() < _toMerge->geoxend() 
			&& geoyroot() < _toMerge->geoyend() )
		{
			return true;
		}
		return false;
	} //}}}

	void GeoData::fuseData(GeoData *_geodata2, GEODATA_BAND _aband, const RasterRect& innerRect, const RasterRect& fusingRect)
	{//{{{
		if (_geodata2 == NULL)
		{
			// stops immediatly
			return;
		}

		// Setting Start index in current geodata
		int iGeo1 = innerRect.yroot()  *            getxsize() + innerRect.xroot();
		// Setting Start index in geodata parameter
		int iGeo2 = fusingRect.yroot() * _geodata2->getxsize() + fusingRect.xroot();
		// Intermediate stops to jump indexes. Rectangles can be in the middle of the Raster.
		int NInter1 = iGeo1 + innerRect.xsize();
		int NInter2 = iGeo2 + fusingRect.xsize();
		// Stop number of pixels to browse
		int NStop1 = innerRect.yend()  * getxsize() + innerRect.xend();
		int NStop2 = fusingRect.yend() * _geodata2->getxsize() + fusingRect.xend();
		double tmp1 = 0.0;
		double tmp2 = 0.0;
		double val  = 0.0;

		// Fusing values that have been converted from images
		switch (fmethod)
		{
			case MEAN:
				while ((iGeo1 <= NStop1) && (iGeo2 <= NStop2))	
				{
					tmp1 = getValueFromBand(_aband, iGeo1);
					tmp2 = _geodata2->getValueFromBand(_aband, iGeo2);
					setValueInBand(_aband, iGeo1, (tmp1+tmp2)/2.0);
					iGeo1++;
					if(iGeo1 == NInter1)
					{
						// Setting iGeo1 to next line start in the defined rect.
						NInter1 = NInter1 + getxsize();
						iGeo1 = NInter1 - innerRect.xsize();
					}
					iGeo2++;
					if(iGeo2 == NInter2)
					{
						// Setting iGeo1 to next line start in the defined rect.
						NInter2 = NInter2 + _geodata2->getxsize();
						iGeo2 = NInter2 - fusingRect.xsize();
					}
				}
				break;
			default: //case BAYES:
				/*--------------------------------
				// Action model // -- // no_action model
				// Sensor model _o for occupied and _f for free
				double mod_o_o = 0.85;
				double mod_o_f = 0.15;

					// apriori
					tmp1 = getValueFromBand(_aband, iGeo1);
					tmp2 = _geodata2->getValueFromBand(_aband, iGeo2);

					if (tmp2 > 0.5) 
					{
						// observation of free zone
						mod_o_o * tmp1

						mod_o_f *
					}
				  --------------------------------- */
				while ((iGeo1 <= NStop1) && (iGeo2 <= NStop2))	
				{
					tmp1 = getValueFromBand(_aband, iGeo1);
					tmp2 = _geodata2->getValueFromBand(_aband, iGeo2);
					// BAYES fusion (from Moravec1 then Burgard1)
					// Taking care of border effects
					if (tmp1 == 1.0) 
					{    
						tmp1 = 0.999;
					}    

					if (tmp1 == 0.0) 
					{    
						tmp1 = 0.001;
					}    

					if (tmp2 == 1.0) 
					{    
						tmp2 = 0.999;
					}    

					if (tmp2 == 0.0) 
					{    
						tmp2 = 0.001;
					}    

					val = (tmp1/ (1.0-tmp1)) * (tmp2/(1.0-tmp2));
					val = (val / (1.0+val));

					setValueInBand(_aband, iGeo1, val);
					iGeo1++;
					if(iGeo1 == NInter1)
					{
						// Setting iGeo1 to next line start in the defined rect.
						NInter1 = NInter1 + getxsize();
						iGeo1 = NInter1 - innerRect.xsize();
					}
					iGeo2++;
					if(iGeo2 == NInter2)
					{
						// Setting iGeo1 to next line start in the defined rect.
						NInter2 = NInter2 + _geodata2->getxsize();
						iGeo2 = NInter2 - fusingRect.xsize();
					}
				}
				break;
		} 

	}//}}}

	double GeoData::calculateBandHomogeneityPercentage(GEODATA_BAND _aband)
	{//{{{
		//GDInfoBand* checkedBand = NULL;
		GDRasterBand<double>* checkedBand = NULL;

		// Selecting the chosen band
		switch(_aband)
		{
			case GEODATA_TROUGH_BAND:
				if (roughTerrainBand == NULL)
				{ return -1.0; }
				checkedBand = roughTerrainBand;
				break;
			case GEODATA_LANDMARK_BAND:
				if (landmarkBand == NULL)
				{ return -1.0; }
				checkedBand = landmarkBand;
				break;
			default:
				// and also case GEODATA_TFLAT_BAND:
				if (traversabilityBand == NULL)
				{ return -1.0; }
				checkedBand = traversabilityBand;
		}

		// heterogeneity counts
		double he_n = 0.0;
		int i, si;
		int j, sj;
		// Value at (i,j)
		double curVal = 0.0;
		// Scanning image ignoring 1 pixel borders
		for (i = 1 ; i < (checkedBand->getxsize() - 1); i++)
		{
			for (j = 1 ; j < (checkedBand->getysize() - 1); j++)
			{
				curVal = checkedBand->getValue(i+j*checkedBand->getxsize());
				// Scanning neighbourHood for heterogeneity
				for(si = i-1; si <= i+1; si++)
				{
					for(sj = j-1; sj <= j+1; sj++)
					{
						if (curVal != checkedBand->getValue(si+sj*checkedBand->getxsize()))
						{
							he_n = he_n + 1.0;
							si = i+2;
							sj = j+2;
						}
					}
				}
			}
		}
		// Normalizing still ignoring 1 pixel borders
		double ho_N = (double)((checkedBand->getxsize() - 2)*(checkedBand->getysize() - 2));
		if (ho_N > 0.0)
		{
			he_n = he_n / ho_N; // value between 0.0 and 1.0
		} else {
			return -1.0;
		} 

		// Inverting to get homogeneity percentage
		he_n = 100.0 * (1.0 - he_n);

		return he_n;

	}//}}}

	bool GeoData::loadVisibility ( const std::string& _visibilityFileName )
	{/*{{{*/
		isInVirtualMemory = 0;
		// Register all GDAL Drivers
		GDALAllRegister();

		if (_visibilityFileName.empty())
		{
			std::cout << "#EEE# [GeoData] Empty file path for visibility map." << std::endl;
			return(false);
		} else {
			loadVisibilityData(_visibilityFileName, NULL);
		}
		return(true);

	 }/*}}}*/

	bool GeoData::loadVisibilityData(const std::string& _visibilityFileName, GDALDataset* _useThisDataset)
	{ /*{{{*/
		//--- Open and read data 
		// Create the GDAL Dataset
		//dataset = (GDALDataset *) GDALOpen( _visibilityFileName.c_str(), GA_ReadOnly );
		if ( _useThisDataset == NULL)
		{
			// Opening image source file
			dataset = (GDALDataset *) GDALOpen( _visibilityFileName.c_str(), GA_Update );
		} else {
			dataset = _useThisDataset;
		}

		// Exit if the GDAL Dataset could not be created
		if( dataset == NULL ) {
			std::cout << "#EEE# [GeoData] Read data file (visibility) error." << std::endl;
			return false;
		}

		// Initializing affine transformation to default transformation
		refGeoTransform[0] = 0.0;
		refGeoTransform[1] = 1.0; // Width
		refGeoTransform[2] = 0.0;
		refGeoTransform[3] = 0.0;
		refGeoTransform[4] = 0.0;
		refGeoTransform[5] = 1.0; // Height (could be -1.0 in some case)

		std::string projectionRef = std::string( dataset->GetProjectionRef());
		if ( projectionRef.length() == 0 )
		{
			//#ifdef LGL_VERBOSE
			std::cout << "#WWW# [GeoData::loadData] Could get any projection type. Setting default projection." <<  std::endl;
			//#endif
			// This is no projection. This is a simple raster file.
			// We have to set the affine transformation ourself.
			dataset->SetGeoTransform(refGeoTransform);
		}

		// Update the GeoTransformation information
		if((refGeoTransformRType=dataset->GetGeoTransform( (double*)refGeoTransform )) != CE_None)
		{
			//#ifdef LGL_VERBOSE
			std::cerr << "#WWW# [GeoData::loadVisibilityData] No geo-reference in current dataset" << std::endl;
			//#endif
		}

		//Only the band number 2 (green) is used
		if (dataset->GetRasterCount() >= 3 ) {
			// Loads the band number GEODATA_VISIBILITY_BAND which is considered as a visibility information band
			// Loads the band number GEODATA_VISIBILITY_HEIGHT_BAND which is considered as height information band for visibility 
			visibilityBand 		 = new GDInfoBand(dataset,dataset->GetRasterBand(3));	// Blue for visibility (the more blue, the more visible)
			visibilityHeightBand = new GDInfoBand(dataset,dataset->GetRasterBand(1));	// Red for Height (the more red, the higher)

		} else if (dataset->GetRasterCount()==1) {
			// Loads the band number GEODATA_VISIBILITY_BAND which is considered as a visibility information band
			visibilityBand 		 = new GDInfoBand(dataset,dataset->GetRasterBand(1)); 	//White for visibility
			// no height information for visibility here
		}

		if (xsize != dataset->GetRasterXSize() || ysize != dataset->GetRasterYSize() ) 
		{
			//#ifdef LGL_VERBOSE
			std::cerr << "#WWW# [GeoData::loadVisibilityData] Dimension between traversability map and visibility map do not match!" << std::endl;
			//#endif
		}

		return true;
	} /*}}}*/

	//---------------------------------------------------------------//

	 void GeoData::printInfo(bool verbose) const 
	 { //{{{
		info(std::cout,verbose);
	 } //}}}

	void GeoData::info(std::ostream& info,bool verbose) const 
	{ /*{{{*/
		//GDALRasterBand *band;
		//int i,iBand;

		/* -------------------------------------------------------------------- */
		/*      Report general info.                                            */
		/* -------------------------------------------------------------------- */
    
		info << "# Geodata file : " << dsfilename << std::endl;

		if (dataset == NULL)
		{
			return;
		}
			  /*
		info << "Driver: " 
			  << dataset->GetDriver()->GetDescription() 
			  << std::endl;
		info << "Driver: " 
			  << dataset->GetDriver()->GetMetadataItem( GDAL_DMD_LONGNAME ) 
			  << std::endl;
			  */
  
		info << "# Size is : xsize="
			  << dataset->GetRasterXSize() 
			  << ",ysize="
			  << dataset->GetRasterYSize()
			  << ",bands="
			  << dataset->GetRasterCount() 
			  << std::endl;

		/* -------------------------------------------------------------------- */
		/*      Report projection.                                              */
		/* -------------------------------------------------------------------- */
		if( dataset->GetProjectionRef()  != NULL )
		  info << "# Projection is: " 
				 << dataset->GetProjectionRef() 
				 << std::endl;


		/* -------------------------------------------------------------------- */
		/*      Report Geotransform.                                            */
		/* -------------------------------------------------------------------- */
		if( refGeoTransformRType == CE_None )
		{
			info << "# Origin : ("
			<< refGeoTransform[0] 
			<< "," 
			<< refGeoTransform[3] 
			<< ")"
			<< std::endl;

			info << "# Pixel Size : (" 
			<< refGeoTransform[1] 
			<< ","
			<< refGeoTransform[5] 
			<< ")"
			<< std::endl;
		} else {
			info << "# [WWW] Could not get a valid GeoTransform. Classical transformation is used (0,1,0,0,0,1)" << std::endl;
		}

		/* -------------------------------------------------------------------- */
		/*      Report GCPs.                                                    */
		/* -------------------------------------------------------------------- */
		/*
		if( dataset->GetGCPCount() > 0 )
		  {
			 printf( "GCP Projection = %s\n", dataset->GetGCPProjection() );
			 for( i = 0; i < dataset->GetGCPCount(); i++ )
				{
				  const GDAL_GCP	*psGCP;
            
				  psGCP = dataset->GetGCPs() + i;
			 
				  std::cout << "GCP[" << i << "]: Id=" <<  psGCP->pszId << ", Info=" << psGCP->pszInfo << std::endl << "(" << psGCP->dfGCPPixel <<","<<  psGCP->dfGCPLine <<") -> ("<< psGCP->dfGCPX <<","<< 	psGCP->dfGCPY <<","<< psGCP->dfGCPZ <<")" << std::endl;	  
				}
		  }
		  */
		//char **papszMetadata;

		/* -------------------------------------------------------------------- */
		/*      Report metadata.                                                */
		/* -------------------------------------------------------------------- */

		/* -------------------------------------------------------------------- */
		/*      Report "IMAGE_STRUCTURE" metadata.                              */
		/* -------------------------------------------------------------------- */

		/* -------------------------------------------------------------------- */
		/*      Report subdatasets.                                             */
		/* -------------------------------------------------------------------- */ 

		/* -------------------------------------------------------------------- */
		/*      Report geolocation.                                             */
		/* -------------------------------------------------------------------- */

	/*
		papszMetadata = dataset->GetDriver()->GetMetadata("GEOLOCATION" );
		if( CSLCount(papszMetadata) > 0 )
		  {
			 printf( "Geolocation:\n" );
			 for( i = 0; papszMetadata[i] != NULL; i++ )
				{
				  printf( "  %s\n", papszMetadata[i] );
				}
		  }
  */
		/* ==================================================================== */
		/*      Loop over bands.                                                */
		/* ==================================================================== */
	/*
		for( iBand = 0; iBand < dataset->GetRasterCount(); iBand++ )
		{
			 double      dfMin, dfMax;// adfCMinMax[2], dfNoData;
			 int         bGotMin, bGotMax; // bGotNodata, bSuccess;
			 int         nBlockXSize, nBlockYSize;// nMaskFlags;
			 //double      dfMean, dfStdDev;
			 //GDALColorTableH	hTable;
			 //CPLErr      eErr;

			 band = dataset->GetRasterBand( iBand+1 );
		
			 // 		double afSample[10000];
			 // 		int   nCount;
			 // 		
			 // 		nCount = band->GetRandomSample( 10000, afSample );
			 // 		info << "Got " << nCount << " samples." << std::endl;
		
			 band->GetBlockSize(  &nBlockXSize, &nBlockYSize );
			 info << "Band " <<iBand+1 << " Block=" << nBlockXSize << "," << nBlockYSize << " Type=" <<  GDALGetDataTypeName(band->GetRasterDataType()) <<" ColorInterp=" << 
				GDALGetColorInterpretationName(band->GetColorInterpretation()) << std::endl;

			 if( band->GetDescription() != NULL 
				  && strlen(band->GetDescription()) > 0 )
				info << "  Description = " << band->GetDescription() << std::endl;
		
			 dfMin = band->GetMinimum( &bGotMin );
			 dfMax = band->GetMaximum( &bGotMax );
			 info << "  ";
			 info << "Min=" << dfMin << " ";
			 info << "Max=" << dfMax << " ";
		
			 // only with new gdal version (1.6.0)
			 // 			 band->ComputeRasterMinMax(FALSE, adfCMinMax );
			 // 			 info << "  Computed Min/Max=" << adfCMinMax[0] << "," << adfCMinMax[1];
		
			 info << std::endl;
			 
		}
		*/
	 } /*}}}*/
	

  }
}
