/**
 * \file LGLViewer.cpp
 *
 * Source file that defines Qt drawing functions
 *
 * \author Redouane Boumghar, Hung Cao
 *
 * \ingroup lgl
 *
 **/

#ifdef HAVE_MODULE_QDISPLAY // lgl viewer is on if HAVE_MODULE_QDISPLAY is defined

#include <lgl/LGLViewer.hpp>
#include <lgl/NavGNode.hpp>

#include <set>

using namespace jafar::qdisplay;

namespace jafar {
	namespace lgl {

typedef void(*FUNCAST)(void*);
		// Default constructor
		LGLViewer::LGLViewer():
		withUTM(false),
		decomp(NULL),
		graph(NULL),
		scaleFactor(1.0),
		n_path_drawn(0),
		visibility(NULL)
		{ //{{{
			// a viewer with screen coordinates (corner 0,0 is at the top left, x is the horizontal axis)
			viewer = new jafar::qdisplay::Viewer();
		} //}}}

		// Specialized constructor
		LGLViewer::LGLViewer(Decomp* _decomp, NavGraph::NavGraphContainer* _graph, double sf):
		withUTM(false),
		decomp(_decomp),
		graph(_graph),
		scaleFactor(sf),
		n_path_drawn(0),
		visibility(NULL)
		{//{{{
			// a viewer with screen coordinates (corner 0,0 is at the top left, x is the horizontal axis)
			viewer = new jafar::qdisplay::Viewer();
		}//}}}

		// Specialized constructor (visibility)
		LGLViewer::LGLViewer(VisibilityMap *_visibility, Decomp* _decomp, NavGraph::NavGraphContainer* _graph, double sf):
		withUTM(false),
		decomp(_decomp),
		graph(_graph),
		scaleFactor(sf),
		n_path_drawn(0),
		visibility(_visibility)
		{//{{{
			// a viewer with screen coordinates (corner 0,0 is at the top left, x is the horizontal axis)
			viewer = new jafar::qdisplay::Viewer();
		}//}}}

		LGLViewer::~LGLViewer()
		{
			delete viewer;
		}

		jafar::qdisplay::Viewer* LGLViewer::getQViewer() {
			return viewer;
		}

		void LGLViewer::setDecomp(Decomp* _decomp)
		{/*{{{*/
			this->decomp = _decomp;
		}/*}}}*/

		void LGLViewer::setGraph(NavGraph::NavGraphContainer* _graph)
		{/*{{{*/
			this->graph = _graph;
		}/*}}}*/

		void LGLViewer::setScaleFactor(double _sf)
		{/*{{{*/
			this->scaleFactor = _sf;
		}/*}}}*/

		void LGLViewer::setImage(const std::string& filename, double xoffset, double yoffset)
		{/*{{{*/
			ImageView* imgV = new ImageView(filename);
			imgV->setPos(xoffset, yoffset);
			imgV->setScale(2.0);
			viewer->setImageView(imgV);
		}/*}}}*/

		void LGLViewer::displayPath(NavPath& foundPath)
		{/*{{{*/
			typedef NavPath::const_iterator path_iterator;
			NavLocation prevPos;
			NavLocation curPos;
			bool drawedger = false;
			for(path_iterator lit = foundPath.begin(); lit != foundPath.end(); ++lit) {
				curPos = (*lit).first;
				displayHook(curPos, DRAW_NODE_PATH);
				if (drawedger)
				{
					drawEdge(prevPos, curPos, DRAW_NODE_PATH);
				} else { drawedger = true; }
				prevPos = curPos;
			}
			n_path_drawn++;
			if (n_path_drawn >= NMAX_PATHS)
			{
				n_path_drawn = 0;
			}
		}/*}}}*/


		bool LGLViewer::displayDecomp()
		{ //{{{

			// In case the LGL Viewer was constructed without "decomp"
			if (decomp == NULL)
			{
				return false;
			}

			// Init.
			int xroot = 0;
			int yroot = 0;
			// Parameters for qdisplay have to be in double precision
			double xcenter = 0;
			double ycenter = 0;
			double xsize = 0;
			double ysize = 0;

			// Show the quadtree decomposition
			Shape* curCluster;
			Decomp_iterator it(decomp);		
			// ++it only increments on positive sized clusters
			for(it.begin(); !(it.at_end()); ++it) {
				// Display this "it" node or leaf from the quadtree decomp.
				it.getIndex(xroot,yroot);

				// For comprehension: we draw the decomposition in Navigation coordinates reference
				// so we introduce (*2.0) factor to translate position from the decomposition raster frame.
				// into the position manager (navigation) frame
				xsize = (double)((*it)->xsize)*2.0*scaleFactor;
				ysize = (double)((*it)->ysize)*2.0*scaleFactor;		
				xcenter = ((double)xroot)*2.0*scaleFactor + xsize/2.0;
				ycenter = ((double)yroot)*2.0*scaleFactor + ysize/2.0;

				curCluster = new Shape(Shape::ShapeRectangle, xcenter, ycenter, xsize, ysize);
				curCluster->setColor(30,50,70);
				viewer->addShape(curCluster);
			}

			return true;
		} //}}}
		
		bool LGLViewer::displayGraph(bool show_edges, bool show_nodes)
		{ /*{{{*/

			if (graph == NULL)
			{
				return false;
			}

			// print the graph		
			NavGraph::NavGraphContainer::GraphContainer* g = graph->getGCtn();

			//--- print all edges
			if (show_edges)
			{
				Line* curEdge;
				NavGraph::NavGraphContainer::Edge e;
				BGL_FORALL_EDGES(e, *g, NavGraph::NavGraphContainer::GraphContainer) {
					curEdge = new Line(
							   (*g)[source(e,*g)].pos.pos_x * scaleFactor,
							   (*g)[source(e,*g)].pos.pos_y * scaleFactor,
							   (*g)[target(e,*g)].pos.pos_x * scaleFactor,
							   (*g)[target(e,*g)].pos.pos_y * scaleFactor);
					curEdge->setColor(120,120,255);
					viewer->addLine(curEdge);
				}
			}

			//--- print all vertices
			if (show_nodes)
			{
				Shape *curVertex;
				//NavGraph::NavGraphContainer::Vertex v;
				BGL_FORALL_VERTICES(v,*g, adjacency_list<>) {
					curVertex = new Shape(Shape::ShapeCross,
								(*g)[v].pos.pos_x * scaleFactor,
								(*g)[v].pos.pos_y * scaleFactor,
								0.6,0.6);
					curVertex->setColor(255,40,40);
					viewer->addShape(curVertex);
				}
			}

			return true;
		} /*}}}*/

		bool LGLViewer::displayVisibilityMap(const int x, const int y, bool show_viewlines, bool show_nodes, bool show_rays)
		{ /*{{{*/

			if (visibility == NULL)
				return false;

			// get the data from the visibilityMap
			const VisibleArea*  myView = visibility->getPtVisibility(x,y) ;
			int xx, yy, ys, vl, n ;
			double alpha;
			ys = myView->getysize() ;
			xx = myView->getOriginalVertex() / ys ; 
			yy = myView->getOriginalVertex() - ( (myView->getOriginalVertex()/ys) * ys ) ;

			vl 		= visibility->getviewLength();
			n 		= visibility->getn();
			alpha 	= visibility->getalpha();
			
			// rounded central node
			Shape *roundedCentralVertex;
			roundedCentralVertex = new Shape(Shape::ShapeRectangleFilled, 	
					xx * scaleFactor,
					yy * scaleFactor,
					2.0,2.0);
			roundedCentralVertex->setColor(230,50,0);
			viewer->addShape(roundedCentralVertex);
			
			// central node
			Shape *centralVertex;
			centralVertex = new Shape(Shape::ShapeRectangleFilled, 	
					x * scaleFactor,
					y * scaleFactor,
					0.5,0.5);
			centralVertex->setColor(255,150,0);
			viewer->addShape(centralVertex);

			//for each node visible from the central node
			std::set<Cell_ID>::const_iterator it;
			for ( it = myView->visibleArea.begin(); it != myView->visibleArea.end() ; ++it )
			{//{{{
				if (show_viewlines) 
				{
					Line* curEdge;
					curEdge = new Line(
						   xx * scaleFactor,
						   yy * scaleFactor,
						   (*it / ys) * scaleFactor,
						   (*it - (*it / ys) * ys) * scaleFactor);
					curEdge->setColor(80,130,100);
					viewer->addLine(curEdge);
				}

				if(show_nodes)
				{
					Shape *curVertex;
					curVertex = new Shape(Shape::ShapeCross,
						    (*it / ys) * scaleFactor,
						   	(*it - (*it / ys) * ys) * scaleFactor,
							0.6,0.6);
					curVertex->setColor(50,250,50);
					viewer->addShape(curVertex);

				}

			}//}}}

			// viewLength circle
			Shape *vCircle;
			vCircle= new Shape(Shape::ShapeEllipse, 
					xx * scaleFactor,
					yy * scaleFactor,
					4*vl,4*vl );
			vCircle->setColor(255,150,0);
			viewer->addShape(vCircle);

			// rays
			if (show_rays)
			{
				// for each ray
				for (int i=0 ; i< n ; i++ )
				{
					Line* curRay;
					curRay = new Line(
						   xx * scaleFactor,
						   yy * scaleFactor,
						   ( xx + cos ( i * alpha ) * 2 * vl ) * scaleFactor,
						   ( yy + sin ( i * alpha ) * 2 * vl ) * scaleFactor);
					curRay->setColor(255,150,0);
					viewer->addLine(curRay);
				}

			}

			return true;
		} /*}}}*/

		bool LGLViewer::displayVisibleArea(const VisibleArea &myArea, bool show_viewlines, bool show_nodes, bool show_rays)
		{ /*{{{*/

			if (visibility == NULL)
				return false;

			// get the data from the visibilityMap
			int xx, yy, ys, vl, n ;
			double alpha;
			ys = myArea.getysize() ;
			xx = myArea.getOriginalVertex() / ys ; 
			yy = myArea.getOriginalVertex() - ( (myArea.getOriginalVertex()/ys) * ys ) ;

			vl 		= visibility->getviewLength();
			n 		= visibility->getn();
			alpha 	= visibility->getalpha();
			
			// rounded central node
			Shape *roundedCentralVertex;
			roundedCentralVertex = new Shape(Shape::ShapeRectangleFilled, 	
					xx * scaleFactor,
					yy * scaleFactor,
					2.0,2.0);
			roundedCentralVertex->setColor(230,50,0);
			viewer->addShape(roundedCentralVertex);
			
			//for each node visible from the central node
			std::set<Cell_ID>::const_iterator it;
			for ( it = myArea.visibleArea.begin(); it != myArea.visibleArea.end() ; ++it )
			{//{{{
				if (show_viewlines) 
				{
					Line* curEdge;
					curEdge = new Line(
						   xx * scaleFactor,
						   yy * scaleFactor,
						   (*it / ys) * scaleFactor,
						   (*it - (*it / ys) * ys) * scaleFactor);
					curEdge->setColor(80,130,100);
					viewer->addLine(curEdge);
				}

				if(show_nodes)
				{
					Shape *curVertex;
					curVertex = new Shape(Shape::ShapeCross,
						    (*it / ys) * scaleFactor,
						   	(*it - (*it / ys) * ys) * scaleFactor,
							0.6,0.6);
					curVertex->setColor(50,250,50);
					viewer->addShape(curVertex);

				}

			}//}}}

			// viewLength circle
			Shape *vCircle;
			vCircle= new Shape(Shape::ShapeEllipse, 
					xx * scaleFactor,
					yy * scaleFactor,
					4*vl,4*vl );
			vCircle->setColor(255,150,0);
			viewer->addShape(vCircle);

			// rays
			if (show_rays)
			{
				// for each ray
				for (int i=0 ; i< n ; i++ )
				{
					Line* curRay;
					curRay = new Line(
						   xx * scaleFactor,
						   yy * scaleFactor,
						   ( xx + cos ( i * alpha ) * 2 * vl ) * scaleFactor,
						   ( yy + sin ( i * alpha ) * 2 * vl ) * scaleFactor);
					curRay->setColor(255,150,0);
					viewer->addLine(curRay);
				}

			}

			return true;
		} /*}}}*/


		bool LGLViewer::displayGraphWithUTM(bool show_edges, bool show_nodes)
		{ /*{{{*/
			if (graph == NULL)
			{
				return false;
			}

			// Telling other functions that we are using UTM
			withUTM = true;

			// Displaying the graph		
			NavGraph::NavGraphContainer::GraphContainer* g = graph->getGCtn();

			//--- print all edges
			if (show_edges)
			{
				Line* curEdge;
				BGL_FORALL_EDGES(e, *g, NavGraph::NavGraphContainer::GraphContainer) {
					curEdge = new Line(
							   (*g)[source(e,*g)].pos.utm_x,
							   (*g)[source(e,*g)].pos.utm_y,
							   (*g)[target(e,*g)].pos.utm_x,
							   (*g)[target(e,*g)].pos.utm_y);
					curEdge->setColor(0,102,255);
					viewer->addLine(curEdge);
				}
			}

			//--- print all vertices
			if (show_nodes)
			{
				Shape *curVertex;
				BGL_FORALL_VERTICES(v,*g, adjacency_list<>) {
					curVertex = new Shape(Shape::ShapeCross,
								(*g)[v].pos.utm_x,
								(*g)[v].pos.utm_y,
								0.6,0.6);
					curVertex->setColor(255,0,0);
					viewer->addShape(curVertex);
				}
			}

			return true;
		} /*}}}*/

		void LGLViewer::drawEdge(const NavGNode& prevNode, const NavGNode& curNode, DRAW_NODE_STATE dstate)
		{//{{{
			if ((&curNode) == NULL || (&prevNode) == NULL)
			{
				return;
			}

			drawEdge(prevNode.pos, curNode.pos, dstate);
		}//}}}

		void LGLViewer::drawEdge(const NavLocation& prevPos, const NavLocation& curPos, DRAW_NODE_STATE dstate)
		{//{{{
			int lowcolor  = 0;
			int highcolor = 0;
			Line *curLine;

			if ( withUTM )
			{
				curLine = new Line(
							  prevPos.utm_x,
							  prevPos.utm_y,
							  curPos.utm_x,
							  curPos.utm_y);
			} else {
				curLine = new Line(
				                  prevPos.pos_x * scaleFactor,
								  prevPos.pos_y * scaleFactor,
				                  curPos.pos_x * scaleFactor,
								  curPos.pos_y * scaleFactor);
			}

			switch(dstate) {
				case DRAW_NODE_PATH: // examine vertex is in the guessed path
					lowcolor  = 40;
					highcolor = 200;
					// Changing color at each new path from green to orange
					highcolor -= n_path_drawn *  (highcolor / NMAX_PATHS); 
					lowcolor  += n_path_drawn * ((255-lowcolor) / NMAX_PATHS);
					// Alternative pick of color 
					if ( (n_path_drawn/2.0) == (double)((int)n_path_drawn/2))
					{
						curLine->setColor(lowcolor, lowcolor, highcolor);
					} else {
						curLine->setColor(highcolor, lowcolor, lowcolor);
					}
					break;
				case DRAW_NODE_TRACK: // examine vertex is in the guessed path
					curLine->setColor(250,250,80);
					break;
				default:
					curLine->setColor(0,0,0);
			}

			if (curLine != NULL)
			{
				viewer->addLine(curLine);
			}

		}//}}}

		void LGLViewer::displayHook(const NavGNode& curNode, DRAW_NODE_STATE dstate)
		{//{{{

			if ((&curNode) == NULL)
			{
				return;
			}

			displayHook(curNode.pos, dstate);

		}//}}}

		void LGLViewer::displayHook(const NavLocation& curPos, DRAW_NODE_STATE dstate)
		{/*{{{*/
			Shape *curBack;
			Shape *curShape;

			if ( withUTM )
			{
				curBack = new Shape(Shape::ShapeEllipseFilled,
								  curPos.utm_x,
				                  curPos.utm_y,
								  0.4, 0.4);
				curShape = new Shape(Shape::ShapeEllipseFilled,
								  curPos.utm_x,
				                  curPos.utm_y,
								  0.3, 0.3);
			} else {
				curBack = new Shape(Shape::ShapeEllipseFilled, 
				                  curPos.pos_x * scaleFactor,
								  curPos.pos_y * scaleFactor,
								  0.5, 0.5);
				curShape = new Shape(Shape::ShapeEllipseFilled, 
				                  curPos.pos_x * scaleFactor,
								  curPos.pos_y * scaleFactor,
								  0.4, 0.4);
			}

			switch(dstate) {
				case DRAW_NODE_OPEN: // discover_vertex has just been added to the OPEN List
					curBack->setColor(40,250,40);
					curShape->setColor(40,40,40);
					break;
				case DRAW_NODE_VISITED: // examine vertex has just been popped out the OPEN List
					curBack->setColor(50,50,250);
					curShape->setColor(250,190,40);
					break;
				case DRAW_NODE_PATH: // examine vertex is in the guessed path
					curBack->setColor(50,50,10);
					curShape->setColor(200,255,200);
					break;
				case DRAW_NODE_TRACK: // examine vertex is in the guessed path
					curBack->setColor(200,200,200);
					curShape->setColor(250,250,80);
					break;
				default:
					curShape->setColor(0,0,0);
			}

			if (curShape != NULL)
			{
				viewer->addShape(curBack);
				viewer->addShape(curShape);
			}

		}/*}}}*/

	}
}

#endif
